###########################################################################
#                                                                         #
#    physical_validation,                                                 #
#    a python package to test the physical validity of MD results         #
#                                                                         #
#    Written by Michael R. Shirts <michael.shirts@colorado.edu>           #
#               Pascal T. Merz <pascal.merz@colorado.edu>                 #
#                                                                         #
#    Copyright (C) 2012 University of Virginia                            #
#              (C) 2017 University of Colorado Boulder                    #
#                                                                         #
#    This library is free software; you can redistribute it and/or        #
#    modify it under the terms of the GNU Lesser General Public           #
#    License as published by the Free Software Foundation; either         #
#    version 2.1 of the License, or (at your option) any later version.   #
#                                                                         #
#    This library is distributed in the hope that it will be useful,      #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of       #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    #
#    Lesser General Public License for more details.                      #
#                                                                         #
#    You should have received a copy of the GNU Lesser General Public     #
#    License along with this library; if not, write to the                #
#    Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,     #
#    Boston, MA 02110-1301 USA                                            #
#                                                                         #
###########################################################################
r"""
This file reimplements most functionality of the checkensemble.py code
originally published on https://github.com/shirtsgroup/checkensemble. It
serves as the low-level functionality of the high-level module
:mod:`physical_validation.ensemble`.
"""
from __future__ import division
import numpy as np
import scipy.optimize

import pymbar

from . import trajectory
from . import error as pv_error
from . import plot


# Pymbar v4 to v3 interface adapter
try:
    pymbar.BAR
except AttributeError:
    pymbar.BAR = pymbar.bar
try:
    pymbar.timeseries.statisticalInefficiency
except AttributeError:
    pymbar.timeseries.statisticalInefficiency = pymbar.timeseries.statistical_inefficiency


def generate_histograms(traj1, traj2, g1, g2, bins):

    n1 = np.size(traj1)
    n2 = np.size(traj2)

    h1 = np.histogram(traj1, bins=bins)[0]/n1
    h2 = np.histogram(traj2, bins=bins)[0]/n2
    dh1 = np.sqrt(g1 * h1 * (1 - h1) / n1)
    dh2 = np.sqrt(g2 * h2 * (1 - h2) / n2)

    return h1, h2, dh1, dh2


def do_linear_fit(traj1, traj2, g1, g2, bins,
                  screen=False, filename=None,
                  trueslope=0.0, trueoffset=0.0,
                  units=None):

    h1, h2, dh1, dh2 = generate_histograms(traj1, traj2, g1, g2, bins)

    #  v  copied from checkensemble.py  v
    ratio = np.log(h2 / h1)
    dratio = np.sqrt((dh1/h1)**2 + (dh2/h2)**2)

    usedat = np.isfinite(ratio)
    y = ratio[usedat]
    nuse = len(y)
    weights = 1.0/dratio[usedat]

    xaxis = (bins[:-1] + bins[1:])/2
    x = xaxis[usedat]

    x_mat = np.ones([nuse, 2])
    x_mat[:, 1] = x

    w = np.diag(weights) 
    wx = np.dot(w, x_mat)
    wy = np.dot(w, y)
    wx_t = np.transpose(wx)
    z = np.dot(wx_t, wx)
    wxy = np.dot(wx_t, wy)

    a = np.linalg.solve(z, wxy)
    da_matrix = np.transpose(np.linalg.inv(z))
    da = np.zeros(2)
    da[0] = np.sqrt(da_matrix[0, 0])
    da[1] = np.sqrt(da_matrix[1, 1])

    # the true line is y = df + dp*x, where y is ln P_1(X)/P_2(X)
    #  ^  end copied from checkensemble.py  ^

    do_plot = screen or filename is not None
    if do_plot:
        true = trueoffset+trueslope*xaxis
        fit = a[0] + a[1]*xaxis

        data = [{'x': xaxis,
                 'y': ratio,
                 'y_err': dratio,
                 'name': 'Simulation'},
                {'x': xaxis,
                 'y': fit,
                 'name': 'Fit to simulation'},
                {'x': xaxis,
                 'y': true,
                 'name': 'Analytical ratio'}]

        if units is not None:
            units = ' [' + units + ']'
        else:
            units = ''

        annot = ('{:.1f}'.format(abs((a[1] - trueslope) / da[1])) +
                 ' quantiles')

        plot.plot(data,
                  legend='best',
                  title='Log probability ratio',
                  xlabel='Energy' + units,
                  ylabel=r'$\log\frac{P_2(E)}{P_1(E)}$',
                  filename=filename,
                  screen=screen,
                  axtext=annot)

    return a, da


def do_max_likelihood_fit(traj1, traj2, g1, g2,
                          init_params=None,
                          verbose=False):

    # ============================================================= #
    # Define (negative) log-likelihood function and its derivatives #
    # ============================================================= #
    def log_likelihood(a, ene1, ene2):
        # Returns negative of eq (8) of check_ensemble paper
        #
        # Uses log (1/f(x)) == -log(f(x))
        # and log(1 + e^x) == log(e^x (e^-x + 1)) == x + log(1 + e^-x)
        #     ^(a)                                   ^(b)
        # form (a) -> 0 for x->-inf, -> inf for x->inf
        # form (b) -> NaN for x->-inf, -> x for x->inf
        # combined: -> 0 for x-> -inf, -> x for x-> inf
        def log_1_plus_exp(y):
            def f(yy):
                with np.errstate(over='raise'):
                    try:
                        xx = np.log(1 + np.exp(yy))
                    except FloatingPointError:
                        xx = yy + np.log(1 + np.exp(-yy))
                    return xx
            return np.vectorize(f)(y)

        if a.size == 2:
            return (np.sum(log_1_plus_exp(a[0] + a[1]*ene1)) +
                    np.sum(log_1_plus_exp(-a[0] - a[1]*ene2)))
        else:
            return (np.sum(log_1_plus_exp(a[0] + a[1]*ene1[0] + a[2]*ene1[1])) +
                    np.sum(log_1_plus_exp(-a[0] - a[1]*ene2[0] - a[2]*ene2[1])))

    def da_log_likelihood(a, ene1, ene2):
        # Returns the first derivative wrt the parameters a of log_likelihood
        #
        # d/da0 log(1 + exp(a0 + a1*E)) == exp(a0 + a1*E) / (1 + exp(a0 + a1*E))
        #                               == 1 / (1 + exp(-a0 - a1*E))
        # d/da1 log(1 + exp(a0 + a1*E)) == E * exp(a0 + a1*E) / (1 + exp(a0 + a1*E))
        #                               == E / (1 + exp(-a0 - a1*E))
        def inv_1_plus_exp(y):
            def f(yy):
                with np.errstate(over='raise'):
                    try:
                        xx = 1. / (1 + np.exp(yy))
                    except FloatingPointError:
                        xx = 0.
                    return xx
            return np.vectorize(f)(y)

        if a.size == 2:
            d = np.zeros(2)
            d[0] = (np.sum(inv_1_plus_exp(-a[0] - a[1]*ene1)) -
                    np.sum(inv_1_plus_exp(a[0] + a[1]*ene2)))
            d[1] = (np.sum(inv_1_plus_exp(-a[0] - a[1]*ene1) * ene1) -
                    np.sum(inv_1_plus_exp(a[0] + a[1]*ene2) * ene2))
        else:
            d = np.zeros(3)
            d[0] = (np.sum(inv_1_plus_exp(-a[0] - a[1]*ene1[0] - a[2]*ene1[1])) -
                    np.sum(inv_1_plus_exp(a[0] + a[1]*ene2[0] + a[2]*ene2[1])))
            d[1] = (np.sum(inv_1_plus_exp(-a[0] - a[1]*ene1[0] - a[2]*ene1[1]) * ene1[0]) -
                    np.sum(inv_1_plus_exp(a[0] + a[1]*ene2[0] + a[2]*ene2[1]) * ene2[0]))
            d[2] = (np.sum(inv_1_plus_exp(-a[0] - a[1]*ene1[0] - a[2]*ene1[1]) * ene1[1]) -
                    np.sum(inv_1_plus_exp(a[0] + a[1]*ene2[0] + a[2]*ene2[1]) * ene2[1]))

        return d

    def hess_log_likelihood(a, ene1, ene2):
        # Returns the hessian wrt the parameters a of log_likelihood
        # fac1 = 1 / (2 + 2*cosh(a0 + a1*ene1))
        # h1 = [[ fac1,      ene1*fac1    ],
        #       [ ene1*fac1, ene1**2*fac1 ]]
        # fac2 = 1 / (2 + 2*cosh(a0 + a1*ene2))
        # h2 = [[ fac2,      ene2*fac2    ],
        #       [ ene2*fac2, ene2**2*fac2 ]]
        # h = h1 + h2

        if a.size == 2:
            fac1 = 1 / (2 + 2*np.cosh(a[0] + a[1]*ene1))
            fac2 = 1 / (2 + 2*np.cosh(a[0] + a[1]*ene2))

            h = np.zeros((2, 2))

            h[0, 0] = np.sum(fac1) + np.sum(fac2)
            h[0, 1] = h[1, 0] = np.sum(ene1 * fac1) + np.sum(ene2 * fac2)
            h[1, 1] = np.sum(ene1 * ene1 * fac1) + np.sum(ene2 * ene2 * fac2)

        else:
            fac1 = 1 / (2 + 2*np.cosh(a[0] + a[1]*ene1[0] + a[2]*ene1[1]))
            fac2 = 1 / (2 + 2*np.cosh(a[0] + a[1]*ene2[0] + a[2]*ene2[1]))

            h = np.zeros((3, 3))

            h[0, 0] = np.sum(fac1) + np.sum(fac2)
            h[1, 1] = np.sum(ene1[0] * ene1[0] * fac1) + np.sum(ene2[0] * ene2[0] * fac2)
            h[2, 2] = np.sum(ene1[1] * ene1[1] * fac1) + np.sum(ene2[1] * ene2[1] * fac2)

            h[0, 1] = h[1, 0] = np.sum(ene1[0] * fac1) + np.sum(ene2[0] * fac2)
            h[0, 2] = h[2, 0] = np.sum(ene1[1] * fac1) + np.sum(ene2[1] * fac2)
            h[1, 2] = h[2, 1] = (np.sum(ene1[0] * ene1[1] * fac1) +
                                 np.sum(ene2[0] * ene2[1] * fac2))

        return h

    # ==================================================== #
    # Minimize the negative of the log likelihood function #
    # ==================================================== #
    if init_params is None:
        init_params = np.zeros(traj1.ndim + 1)
    else:
        init_params = np.array(init_params)

    min_res = scipy.optimize.minimize(
        log_likelihood,
        x0=init_params,
        args=(traj1, traj2),
        method='dogleg',
        jac=da_log_likelihood,
        hess=hess_log_likelihood
    )

    # fallback options
    if not min_res.success:
        if verbose:
            print('Note: Max-Likelihood minimization failed using \'dogleg\' method. '
                  'Trying to vary initial parameters.')
        min_res_1 = scipy.optimize.minimize(
            log_likelihood,
            x0=init_params * 0.9,
            args=(traj1, traj2),
            method='dogleg',
            jac=da_log_likelihood,
            hess=hess_log_likelihood
        )
        min_res_2 = scipy.optimize.minimize(
            log_likelihood,
            x0=init_params * 1.1,
            args=(traj1, traj2),
            method='dogleg',
            jac=da_log_likelihood,
            hess=hess_log_likelihood
        )
        if min_res_1.success and min_res_2.success and np.allclose(min_res_1.x, min_res_2.x):
            min_res = min_res_1

    if not min_res.success:
        # dogleg was unsuccessful using alternative starting point
        if verbose:
            print('Note: Max-Likelihood minimization failed using \'dogleg\' method. '
                  'Trying method \'nelder-mead\'.')
        min_res = scipy.optimize.minimize(
            log_likelihood,
            x0=init_params * 0.9,
            args=(traj1, traj2),
            method='nelder-mead'
        )

    if not min_res.success:
        raise RuntimeError('MaxLikelihood: Unable to minimize function.')

    final_params = min_res.x

    # ======================= #
    # Calculate uncertainties #
    # ======================= #
    cov = np.linalg.inv(hess_log_likelihood(final_params, traj1, traj2))
    final_error = np.sqrt(np.diag(cov))*np.sqrt(np.average([g1, g2]))

    return final_params, final_error


def check_bins(traj1, traj2, bins):
    # check for empty bins
    h1, _ = np.histogram(traj1, bins=bins)
    h2, _ = np.histogram(traj2, bins=bins)
    empty = np.where((h1 == 0) | (h2 == 0))[0]

    if np.size(empty) == 0:
        return bins
    elif np.size(empty) == 1:
        empty = empty[0]
        if empty > np.size(bins) / 2:
            return bins[:empty]
        else:
            return bins[empty+1:]
    else:
        # find longest non-empty interval
        empty = np.insert(np.append(empty, [40]), 0, [-1])
        max_interval = np.argmax(empty[1:] - empty[:-1])
        left = empty[max_interval] + 1
        right = empty[max_interval + 1]
        return bins[left:right]


def print_stats(title,
                fitvals, dfitvals,
                kb, param1, param2, trueslope,
                temp=None, pvconvert=None,
                dtemp=False, dpress=False, dmu=False,
                dtempdpress=False, dtempdmu=False):

    # if simple 1d:
    #     fitvals = [df, slope]
    #     dfitvals = [ddf, dslope]
    # if simple 2d:
    #     fitvals = [df, slope0, slope1]
    #     dfitvals = [ddf, dslope0, dslope1]
    # if bootstrapped 1d:
    #     fitvals = [[df, slope], [df, slope], ...]
    #     dfitvals = None
    # if bootstrapped 2d:
    #     fitvals = [[df, slope0, slope1], [df, slope0, slope1], ...]
    #     dfitvals = None
    if fitvals.ndim > 1:
        dfitvals = np.std(fitvals, axis=0)
        fitvals = np.average(fitvals, axis=0)

    if np.ndim(trueslope) == 0:
        trueslopes = np.array([trueslope])
    else:
        trueslopes = trueslope

    free_energy = fitvals[0]
    slopes = fitvals[1:]
    dfree_energy = dfitvals[0]
    dslopes = dfitvals[1:]

    print('='*50)
    print(title)
    print('='*50)
    print('Free energy')
    print('    {:.5f} +/- {:.5f}'.format(free_energy, dfree_energy))
    print('{:27s}      |  {:s}'.format('Estimated slope', 'True slope'))
    for slope, dslope, trueslope in zip(slopes, dslopes, trueslopes):
        print('    {:<9.6f} +/- {:<9.6f}      |  {:<9.6f}'.format(slope, dslope, trueslope))
        quant = np.abs((slope-trueslope)/dslope)
        print('    ({:.2f} quantiles from true slope)'.format(quant))

    if dtemp or dtempdpress or dtempdmu:
        # slope is estimated beta2 - beta1
        # kb * slope == 1/T1' - 1/T2' == (T2' - T1')/(T1'*T2')
        # So we'll assume dT' == T1' - T2' ~= kb * slope * T1*T2
        slope = slopes[0]
        dslope = dslopes[0]
        if dtemp:
            t1 = param1
            t2 = param2
        else:
            t1 = param1[0]
            t2 = param2[0]
        print('{:27s}      |  {:s}'.format('Estimated dT', 'True dT'))
        print('    {:<6.1f} +/- {:<6.1f}            |  {:<6.1f}'.format(
            kb * slope * t1 * t2,
            kb * dslope * t1 * t2,
            t2 - t1
        ))
    if dpress or dtempdpress:
        # slope is estimated (P1 - P2)/beta*pvconvert (1d), or
        #                    (P1/b1 - P2/b2)*pvconvert (2d)
        if temp is None and dtempdpress:
            temp = .5*(param1[0] + param2[0])
        if dpress:
            press = -slopes[0] * (kb*temp) / pvconvert
            ddpress = -dslopes[0] * (kb*temp) / pvconvert
            truepress = -trueslopes[0] * (kb*temp) / pvconvert
        else:
            press = -slopes[1] * (kb*temp) / pvconvert
            ddpress = -dslopes[1] * (kb*temp) / pvconvert
            truepress = -trueslopes[1] * (kb*temp) / pvconvert
        print('{:27s}      |  {:s}'.format('Estimated dP', 'True dP'))
        print('    {:<6.1f} +/- {:<6.1f}            |  {:<6.1f}'.format(
            press, np.abs(ddpress), truepress
        ))
    if dmu or dtempdmu:
        pass
    print('='*50)


def estimate_interval(ens_string, ens_temp,
                      energy, kb,
                      ens_press=None,
                      volume=None, pvconvert=None,
                      verbosity=1, cutoff=0.001,
                      tunit='', punit=''):
    result = {}
    if ens_string == 'NVT':
        # Discard burn-in period and time-correlated frames
        energy = trajectory.equilibrate(energy, verbose=(verbosity > 1), name='Energy')
        energy = trajectory.decorrelate(energy, verbose=(verbosity > 1), name='Energy')
        energy = trajectory.cut_tails(energy, cut=cutoff, verbose=(verbosity > 2), name='Energy')

        # dT
        sig = np.std(energy)
        result['dT'] = 2*kb*ens_temp*ens_temp/sig
    elif ens_string == 'NPT':
        enthalpy = energy + pvconvert * ens_press * volume
        traj_2d = np.array([energy, volume])
        # Discard burn-in period and time-correlated frames
        enthalpy = trajectory.equilibrate(enthalpy, verbose=(verbosity > 1), name='Enthalpy')
        enthalpy = trajectory.decorrelate(enthalpy, verbose=(verbosity > 1), name='Enthalpy')
        enthalpy = trajectory.cut_tails(enthalpy, cut=cutoff, verbose=(verbosity > 2), name='Enthalpy')
        volume_1d = trajectory.equilibrate(volume, verbose=(verbosity > 1), name='Volume')
        volume_1d = trajectory.decorrelate(volume_1d, verbose=(verbosity > 1), name='Volume')
        volume_1d = trajectory.cut_tails(volume_1d, cut=cutoff, verbose=(verbosity > 2), name='Volume')
        traj_2d = trajectory.equilibrate(traj_2d, verbose=(verbosity > 1), name='2D-Trajectory')
        traj_2d = trajectory.decorrelate(traj_2d, facs=[1, pvconvert * ens_press], verbose=(verbosity > 1), name='2D-Trajectory')
        traj_2d = trajectory.cut_tails(traj_2d, cut=cutoff, verbose=(verbosity > 2), name='2D-Trajectory')

        # dT
        sig = np.std(enthalpy)
        result['dT'] = 2*kb*ens_temp*ens_temp/sig
        # dP
        sig = np.std(volume_1d)*pvconvert
        result['dP'] = 2*kb*ens_temp/sig
        # dTdP
        cov = np.cov(traj_2d)
        sig = np.sqrt(np.diag(cov))
        sig[1] *= pvconvert
        result['dTdP'] = [2*kb*ens_temp*ens_temp/sig[0],
                          2*kb*ens_temp/sig[1]]
    else:
        raise pv_error.InputError('ens_str', 'Unrecognized ensemble string.')

    if verbosity > 0:
        print('A rule of thumb states that good error recognition can be expected when\n'
              'spacing the tip of the distributions by about two standard deviations.\n'
              'Based on this rule, and the assumption that the standard deviation of the\n'
              'distributions is largely independent of the state point, here\'s an estimate\n'
              'for the interval given the current simulation:')
        if ens_string == 'NVT':
            print('Current trajectory: NVT, T = {:.2f} {:s}'.format(ens_temp, tunit))
            print('Suggested interval: dT = {:.1f} {:s}'.format(result['dT'], tunit))
        if ens_string == 'NPT':
            print('Current trajectory: NPT, T = {:.2f} {:s}, P = {:.2f} {:s}'.format(
                ens_temp, tunit, ens_press, punit))
            print('Suggested interval:')
            print('  Temperature-only: dT = {:.1f} {:s}'.format(result['dT'], tunit))
            print('  Pressure-only: dP = {:.1f} {:s}'.format(result['dP'], punit))
            print('  Combined: dT = {:.1f} {:s}, dP = {:.1f} {:s}'.format(
                result['dTdP'][0], tunit, result['dTdP'][1], punit))


def check_1d(traj1, traj2, param1, param2, kb,
             quantity, dtemp=False, dpress=False, dmu=False,
             temp=None, pvconvert=None,
             nbins=40, cutoff=0.001, seed=None,
             verbosity=1, screen=False, filename=None):
    r"""
    Checks whether the energy trajectories of two simulation performed at
    different temperatures have sampled distributions at the analytically
    expected ratio.

    Parameters
    ----------
    traj1 : array-like
        Trajectory of the first simulation
        If dtemp:

            * NVT: Potential energy U or total energy E = U + K
            * NPT: Enthalpy H = U + pV or total energy E = H + K

        If dpress:

            * NPT: Volume V

    traj2 : array-like
        Trajectory of the second simulation
        If dtemp:

            * NVT: Potential energy U or total energy E = U + K
            * NPT: Enthalpy H = U + pV or total energy E = H + K

        If dpress:

            * NPT: Volume V

    param1 : float
        Target temperature or pressure of the first simulation
    param2 : float
        Target temperature or pressure of the second simulation
    kb : float
        Boltzmann constant in same units as the energy trajectories
    quantity : str
        Name of quantity analyzed (used for printing only)
    dtemp : bool, optional
        Set to True if trajectories were simulated at different temperature
        Default: False.
    dpress : bool, optional
        Set to True if trajectories were simulated at different pressure
        Default: False.
    temp : float, optional
        The temperature in equal temperature, differring pressure NPT simulations.
        Needed to print optimal dP.
    pvconvert : float, optional
        Conversion from pressure * volume to energy units.
        Needed to print optimal dP.
    dmu : bool, optional
        Set to True if trajectories were simulated at different chemical potential
        Default: False.
    nbins : int, optional
        Number of bins used to assess distributions of the trajectories
        Default: 40
    cutoff : float, optional
        Tail cutoff of distributions.
        Default: 0.001 (0.1%)
    seed : int, optional
        If set, bootstrapping will be reproducible.
        Default: None, bootstrapping non-reproducible.
    verbosity : int, optional
        Verbosity level.
        Default: 1 (only most important output)
    screen : bool, optional
        Plot distributions on screen.
        Default: False.
    filename : string, optional
        Plot distributions to `filename`.pdf.
        Default: None.

    Returns
    -------

    """

    if (not (dtemp or dpress or dmu) or
       (dtemp and dpress) or
       (dtemp and dmu) or
       (dpress and dmu)):
        raise pv_error.InputError(['dtemp', 'dpress', 'dmu'],
                                  'Need to specify exactly one of `dtemp`, `dpress` and `dmu`.')

    if dmu:
        raise NotImplementedError('check_1d: Testing of `dmu` not implemented.')

    if seed is not None:
        raise NotImplementedError('check_1d: Bootstrapping not implemented.')

    if dpress and (temp is None or pvconvert is None):
        raise pv_error.InputError(['dpress', 'temp', 'pvconvert'],
                                  '`ensemble.check_1d` with `dpress=True` requires `temp` and `pvconvert`.')

    # =============================== #
    # prepare constants, strings etc. #
    # =============================== #
    pstring = 'ln(P_2(' + quantity + ')/P_1(' + quantity + '))'
    trueslope = 0
    if dtemp:
        trueslope = 1/(kb * param1) - 1/(kb * param2)
    elif dpress:
        trueslope = (param1 - param2) / (kb * temp) * pvconvert

    if verbosity > 1:
        print('Analytical slope of {:s}: {:.8f}'.format(pstring, trueslope))

    quant = {}

    # ==================== #
    # prepare trajectories #
    # ==================== #
    # Discard burn-in period and time-correlated frames
    traj1 = trajectory.prepare(traj1, cut=cutoff, verbosity=verbosity, name='Trajectory 1')
    traj2 = trajectory.prepare(traj2, cut=cutoff, verbosity=verbosity, name='Trajectory 2')

    # calculate inefficiency
    g1 = pymbar.timeseries.statisticalInefficiency(traj1)
    g2 = pymbar.timeseries.statisticalInefficiency(traj2)

    # calculate overlap
    traj1_full = traj1
    traj2_full = traj2
    traj1, traj2, min_ene, max_ene = trajectory.overlap(
        traj1=traj1_full, traj2=traj2_full,
    )
    if verbosity > 0:
        print('Overlap is {:.1%} of trajectory 1 and {:.1%} of trajectory 2.'.format(
            traj1.shape[0] / traj1_full.shape[0],
            traj2.shape[0] / traj2_full.shape[0]
        ))
    if verbosity > 0 and dtemp:
        sig1 = np.std(traj1_full)
        sig2 = np.std(traj2_full)
        dt1 = 2*kb*param1*param1/sig1
        dt2 = 2*kb*param2*param2/sig2
        if verbosity > 1:
            print('A rule of thumb states that a good overlap is found when dT/T = (2*kB*T)/(sig),\n'
                  'where sig is the standard deviation of the energy distribution.\n'
                  'For the current trajectories, dT = {:.1f}, sig1 = {:.1f} and sig2 = {:.1f}.\n'
                  'According to the rule of thumb, given T1, a good dT is dT = {:.1f}, and\n'
                  '                                given T2, a good dT is dT = {:.1f}.'.format(
                      param2-param1, sig1, sig2, dt1, dt2)
                  )
        print('Rule of thumb estimates that dT = {:.1f} would be optimal '
              '(currently, dT = {:.1f})'.format(.5*(dt1+dt2), param2-param1))
    if verbosity > 0 and dpress:
        sig1 = np.std(traj1_full)*pvconvert
        sig2 = np.std(traj2_full)*pvconvert
        dp1 = 2*kb*temp/sig1
        dp2 = 2*kb*temp/sig2
        if verbosity > 1:
            print('A rule of thumb states that a good overlap is found when dP = (2*kB*T)/(sig),\n'
                  'where sig is the standard deviation of the volume distribution.\n'
                  'For the current trajectories, dP = {:.1f}, sig1 = {:.1g} and sig2 = {:.1g}.\n'
                  'According to the rule of thumb, given P1, a good dP is dP = {:.1f}, and\n'
                  '                                given P2, a good dP is dP = {:.1f}.'.format(
                      param2-param1, sig1, sig2, dp1, dp2)
                  )
        print('Rule of thumb estimates that dP = {:.1f} would be optimal '
              '(currently, dP = {:.1f})'.format(.5*(dp1+dp2), param2-param1))
    if not min_ene:
        raise pv_error.InputError(['traj1', 'traj2'],
                                  'No overlap between trajectories.')
    # calculate bins
    bins = np.linspace(min_ene, max_ene, nbins+1)
    bins = check_bins(traj1, traj2, bins)
    if np.size(bins) < 3:
        raise pv_error.InputError(['traj1', 'traj2', 'nbins', 'cutoff'],
                                  'Less than 3 bins were filled in the overlap region.\n'
                                  'Ensure sufficient overlap between the trajectories, and '
                                  'consider increasing `cutoff` or `nbins` if there is '
                                  'sufficient overlap but unusually long tails.')

    w_f = -trueslope * traj1_full
    w_r = trueslope * traj2_full

    if verbosity > 2:
        print('Computing log of partition functions using pymbar.BAR...')
    df_ddf = pymbar.BAR(w_f, w_r)
    try:
        # tuple in v3
        df = df_ddf[0]
        ddf = df_ddf[1]
    except KeyError:
        # dict in v4
        df = df_ddf['Delta_f']
        ddf = df_ddf['dDelta_f']
    if verbosity > 2:
        print('Using {:.5f} for log of partition functions as computed from BAR.'.format(df))
        print('Uncertainty in quantity is {:.5f}.'.format(ddf))
        print('Assuming this is negligible compared to sampling error at individual points.')

    # ========== #
    # linear fit #
    # ========== #
    if verbosity > 2:
        print('Computing linear fit parameters (for plotting / comparison)')

    fitvals, dfitvals = do_linear_fit(
        traj1=traj1_full, traj2=traj2_full, g1=g1, g2=g2, bins=bins,
        screen=screen, filename=filename,
        trueslope=trueslope, trueoffset=df,
        units=None
    )

    slope = fitvals[1]
    dslope = dfitvals[1]
    quant['linear'] = [abs((slope - trueslope)/dslope)]
    if verbosity > 1:
        print_stats(
            title='Linear Fit Analysis (analytical error)',
            fitvals=fitvals,
            dfitvals=dfitvals,
            kb=kb,
            param1=param1,
            param2=param2,
            trueslope=trueslope,
            temp=temp, pvconvert=pvconvert,
            dtemp=dtemp, dpress=dpress, dmu=dmu
        )

    # ================== #
    # max-likelihood fit #
    # ================== #
    if verbosity > 2:
        print('Computing the maximum likelihood parameters')

    fitvals, dfitvals = do_max_likelihood_fit(traj1_full, traj2_full, g1, g2,
                                              init_params=[df, trueslope],
                                              verbose=(verbosity > 1))

    slope = fitvals[1]
    dslope = dfitvals[1]
    quant['maxLikelihood'] = [abs((slope - trueslope)/dslope)]
    if verbosity > 0:
        print_stats(
            title='Maximum Likelihood Analysis (analytical error)',
            fitvals=fitvals,
            dfitvals=dfitvals,
            kb=kb,
            param1=param1,
            param2=param2,
            trueslope=trueslope,
            temp=temp, pvconvert=pvconvert,
            dtemp=dtemp, dpress=dpress, dmu=dmu
        )

    return quant['maxLikelihood']


def check_2d(traj1, traj2, param1, param2, kb, pvconvert,
             quantity, dtempdpress=False, dtempdmu=False,
             cutoff=0.001, seed=None,
             verbosity=1, screen=False, filename=None):
    r"""
    Checks whether the energy trajectories of two simulation performed at
    different temperatures have sampled distributions at the analytically
    expected ratio.

    Parameters
    ----------
    traj1 : array-like, 2d
        Trajectory of the first simulation
        If dtempdpress:

            * traj[0,:]: Potential energy U or total energy E = U + K
            * traj[1,:]: Volume V
    traj2 : array-like, 2d
        Trajectory of the second simulation
        If dtempdpress:

            * traj[0,:]: Potential energy U or total energy E = U + K
            * traj[1,:]: Volume V
    param1 : array-like
        If dtempdpress:
            Target temperature and pressure of the first simulation
    param2 : array-like
        If dtempdpress:
            Target temperature and pressure of the first simulation
    kb : float
        Boltzmann constant in same units as the energy trajectories
    pvconvert : float
        Conversion from pressure * volume to energy units
    quantity : List[str]
        Names of quantities analyzed (used for printing only)
    dtempdpress : bool, optional
        Set to True if trajectories were simulated at different
        temperature and pressure
        Default: False.
    dtempdmu : bool, optional
        Set to True if trajectories were simulated at different
        temperature and chemical potential
        Default: False.
    cutoff : float
        Tail cutoff of distributions.
        Default: 0.001 (0.1%)
    seed : int
        If set, bootstrapping will be reproducible.
        Default: None, bootstrapping non-reproducible.
    verbosity : int
        Verbosity level.
        Default: 1 (only most important output)
    screen : bool, optional
        Plot distributions on screen.
        Default: False.
    filename : string, optional
        Plot distributions to `filename`.pdf.
        Default: None.

    Returns
    -------

    """

    if not (dtempdpress or dtempdmu) or (dtempdpress and dtempdmu):
        raise pv_error.InputError(['dtempdpress', 'dtempdmu'],
                                  'Need to specify exactly one of `dtempdpress` and `dtempdmu`.')

    if dtempdmu:
        raise NotImplementedError('check_2d: Testing of `dtempdmu` not implemented.')

    if seed is not None:
        raise NotImplementedError('check_2d: Bootstrapping not implemented.')

    if screen or filename is not None:
        raise NotImplementedError('check_2d: Plotting not implemented.')

    # =============================== #
    # prepare constants, strings etc. #
    # =============================== #
    pstring = ('ln(P_2(' + quantity[0] + ', ' + quantity[1] + ')/' +
               'P_1(' + quantity[0] + ', ' + quantity[1] + '))')
    trueslope = np.zeros(2)
    facs = [None, None]
    if dtempdpress:
        trueslope = np.array([
            1/(kb * param1[0]) - 1/(kb * param2[0]),
            pvconvert*(1/(kb * param1[0]) * param1[1] - 1/(kb * param2[0]) * param2[1])
        ])
        facs = [[1, param1[1]], [1, param2[1]]]

    if verbosity > 1:
        print('Analytical slope of {:s}: {:.8f}, {:.8f}'.format(
            pstring, trueslope[0], trueslope[1]
        ))

    quant = {}

    # ==================== #
    # prepare trajectories #
    # ==================== #
    # Discard burn-in period and time-correlated frames
    traj1 = trajectory.prepare(traj1, cut=cutoff, facs=facs[0],
                               verbosity=verbosity, name='Trajectory 1')
    traj2 = trajectory.prepare(traj2, cut=cutoff, facs=facs[1],
                               verbosity=verbosity, name='Trajectory 2')

    # calculate inefficiency
    g1 = np.array([
            pymbar.timeseries.statisticalInefficiency(traj1[0]),
            pymbar.timeseries.statisticalInefficiency(traj1[1])
        ])
    g2 = np.array([
            pymbar.timeseries.statisticalInefficiency(traj2[0]),
            pymbar.timeseries.statisticalInefficiency(traj2[1])
        ])

    # calculate overlap
    traj1_full = traj1
    traj2_full = traj2
    traj1, traj2, min_ene, max_ene = trajectory.overlap(
        traj1=traj1_full, traj2=traj2_full,
    )
    if verbosity > 0:
        print('Overlap is {:.1%} of trajectory 1 and {:.1%} of trajectory 2.'.format(
            traj1.shape[1] / traj1_full.shape[1],
            traj2.shape[1] / traj2_full.shape[1]
        ))
    if verbosity > 0 and dtempdpress:
        cov1 = np.cov(traj1_full)
        sig1 = np.sqrt(np.diag(cov1))
        sig1[1] *= pvconvert
        cov2 = np.cov(traj2_full)
        sig2 = np.sqrt(np.diag(cov2))
        sig2[1] *= pvconvert
        dt1 = 2*kb*param1[0]*param1[0]/sig1[0]
        dt2 = 2*kb*param2[0]*param2[0]/sig2[0]
        dp1 = 2*kb*param1[0]/sig1[1]
        dp2 = 2*kb*param2[0]/sig2[1]
        if verbosity > 1:
            print('A rule of thumb states that a good overlap can be expected when choosing state\n'
                  'points separated by about 2 standard deviations.\n'
                  'For the current trajectories, dT = {:.1f}, and dP = {:.1f},\n'
                  'with standard deviations sig1 = [{:.1f}, {:.1g}], and sig2 = [{:.1f}, {:.1g}].\n'
                  'According to the rule of thumb, given point 1, the estimate is dT = {:.1f}, dP = {:.1f}, and\n'
                  '                                given point 2, the estimate is dT = {:.1f}, dP = {:.1f}.'.format(
                      param2[0]-param1[0], param2[1]-param1[1],
                      sig1[0], sig1[1], sig2[0], sig2[1],
                      dt1, dt2, dp1, dp2)
                  )
        print('Rule of thumb estimates that (dT,dP) = ({:.1f},{:.1f}) would be optimal '
              '(currently, (dT,dP) = ({:.1f},{:.1f}))'.format(.5*(dt1+dt2), .5*(dp1+dp2),
                                                              param2[0]-param1[0], param2[1]-param1[1]))
    if min_ene is None:
        raise pv_error.InputError(['traj1', 'traj2'],
                                  'No overlap between trajectories.')

    w_f = -trueslope[0] * traj1_full[0] - trueslope[1] * traj1_full[1]
    w_r = trueslope[0] * traj2_full[0] + trueslope[1] * traj2_full[1]

    if verbosity > 2:
        print('Computing log of partition functions using pymbar.BAR...')
    df_ddf = pymbar.BAR(w_f, w_r)
    try:
        # tuple in v3
        df = df_ddf[0]
        ddf = df_ddf[1]
    except KeyError:
        # dict in v4
        df = df_ddf['Delta_f']
        ddf = df_ddf['dDelta_f']
    if verbosity > 2:
        print('Using {:.5f} for log of partition functions as computed from BAR.'.format(df))
        print('Uncertainty in quantity is {:.5f}.'.format(ddf))
        print('Assuming this is negligible compared to sampling error at individual points.')

    # ================== #
    # max-likelihood fit #
    # ================== #
    if verbosity > 2:
        print('Computing the maximum likelihood parameters')

    fitvals, dfitvals = do_max_likelihood_fit(traj1_full, traj2_full, g1, g2,
                                              init_params=[df, trueslope[0], trueslope[1]],
                                              verbose=(verbosity > 1))

    slope = fitvals[1:]
    dslope = dfitvals[1:]
    quant['maxLikelihood'] = np.abs((slope - trueslope)/dslope)
    if verbosity > 0:
        print_stats(
            title='Maximum Likelihood Analysis (analytical error)',
            fitvals=fitvals,
            dfitvals=dfitvals,
            kb=kb,
            param1=param1,
            param2=param2,
            trueslope=trueslope,
            pvconvert=pvconvert,
            dtempdpress=dtempdpress, dtempdmu=dtempdmu
        )

    return quant['maxLikelihood']



#!/usr/bin/env python3
#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2014- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

"""Check source code and Doxygen documentation for issues

This script checks for some issues in the Doxygen documentation, as well as
general issues in the source code, mainly using Doxygen XML output and #include
dependencies parsed from source files.  Part of the checks are generic, like
checking that all documented entities have brief descriptions.

The checks should be self-evident from the source code of the script
(they are also described in docs/dev-manual/gmxtree.rst).
All the logic of parsing the Doxygen XML output and creating a GROMACS-specific
representation of the source tree is separated into separate Python modules
(doxygenxml.py and gmxtree.py, respectively).  Similarly, logic for handling
the output messages is in reporter.py.   This leaves only the actual checks and
the script command-line interface in this file.

The script can be run using the 'check-source' target generated by CMake.
This target takes care of generating all the necessary input files and passing
them to the script.
"""

import sys
from optparse import OptionParser

import gmxtree
from gmxtree import GromacsTree, DocType
from reporter import Reporter


def check_file(fileobj, tree, reporter):
    """Check file-level issues."""
    if not fileobj.is_external() and fileobj.get_relpath().startswith("src/"):
        includes = fileobj.get_includes()
        if fileobj.is_source_file():
            if includes:
                firstinclude = includes[0].get_file()
                if not firstinclude or firstinclude.get_name() != "gmxpre.h":
                    reporter.code_issue(
                        includes[0], 'does not include "gmxpre.h" first'
                    )
            else:
                reporter.code_issue(fileobj, 'does not include "gmxpre.h"')
        used_define_files = fileobj.get_used_define_files()
        for define_file in tree.get_checked_define_files():
            includes_file = False
            for include in includes:
                if include.get_file() == define_file:
                    includes_file = True
                    break
            if includes_file:
                if not define_file in used_define_files:
                    reporter.code_issue(
                        fileobj,
                        'includes "{0}" unnecessarily'.format(define_file.get_name()),
                    )
            else:
                if define_file in used_define_files:
                    used_defines = list(fileobj.get_used_defines(define_file))
                    if len(used_defines) > 3:
                        used_defines = used_defines[:3] + ["..."]
                    used_defines = ", ".join(used_defines)
                    reporter.code_issue(
                        fileobj,
                        'should include "{0}"'.format(define_file.get_name()),
                        details=["uses " + used_defines],
                    )

    if not fileobj.is_documented():
        # TODO: Add rules for required documentation
        return

    if fileobj.is_source_file():
        # TODO: Add rule to exclude examples from this check
        if fileobj.get_doc_type() != DocType.internal:
            reporter.file_error(
                fileobj, "source file documentation appears outside full documentation"
            )
        elif fileobj.get_api_type() != DocType.internal:
            reporter.file_error(fileobj, "source file marked as non-internal")
        elif fileobj.get_doc_type() < fileobj.get_api_type():
            reporter.file_error(
                fileobj,
                "API type ({0}) conflicts with documentation visibility ({1})".format(
                    fileobj.get_api_type(), fileobj.get_doc_type()
                ),
            )

    if not fileobj.has_brief_description():
        reporter.file_error(
            fileobj, "is documented, but does not have brief description"
        )

    expectedmod = fileobj.get_expected_module()
    if expectedmod:
        docmodules = fileobj.get_doc_modules()
        if docmodules:
            for module in docmodules:
                if module != expectedmod:
                    reporter.file_error(
                        fileobj,
                        "is documented in incorrect module: {0}".format(
                            module.get_name()
                        ),
                    )
        elif expectedmod.is_documented():
            reporter.file_error(
                fileobj,
                "is not documented in any module, but {0} exists".format(
                    expectedmod.get_name()
                ),
            )


def check_include(fileobj, includedfile, reporter):
    """Check an #include directive."""
    otherfile = includedfile.get_file()
    if includedfile.is_system():
        if not otherfile:
            return
        reporter.code_issue(
            includedfile, "includes local file as {0}".format(includedfile)
        )
    # To report compiler flags in the log file, include files must be
    # generated by CMake, with a name that is only known at generation
    # time (not even configuration time). This looks like the
    # inclusion of a non-local file to the source checker. We usually
    # suppress issues with generated files using file-name matching,
    # but this is not suitable in this case because it is hard to map
    # the repository file named src/compilerflaginfo.h.cmakein to two
    # generated header files
    # $builddir/src/compilerflaginfo-$buildtype-CXX.h and
    # $builddir/src/compilerflaginfo-$buildtype-C.h. So we use a hack
    # to just avoid the warning for this special source file.
    if "compilerflaginfo" in str(includedfile):
        return
    if not otherfile:
        reporter.code_issue(
            includedfile, "includes non-local file as {0}".format(includedfile)
        )
    if not otherfile:
        return
    filemodule = fileobj.get_module()
    othermodule = otherfile.get_module()
    if fileobj.is_documented() and otherfile.is_documented():
        filetype = fileobj.get_doc_type()
        othertype = otherfile.get_doc_type()
        if filetype > othertype:
            reporter.code_issue(
                includedfile,
                "{0} file includes {1} file {2}".format(
                    filetype, othertype, includedfile
                ),
            )
    check_api = otherfile.api_type_is_reliable() and filemodule != othermodule


def check_entity(entity, reporter):
    """Check documentation for a code construct."""
    if entity.is_documented():
        if not entity.has_brief_description():
            reporter.doc_error(
                entity, "is documented, but does not have brief description"
            )


def check_class(classobj, reporter):
    """Check documentation for a class/struct/union."""
    check_entity(classobj, reporter)
    if classobj.is_documented():
        classtype = classobj.get_doc_type()
        filetype = classobj.get_file_doc_type()
        if filetype is not DocType.none and classtype > filetype:
            reporter.doc_error(
                classobj,
                "is in {0} file(s), but appears in {1} documentation".format(
                    filetype, classtype
                ),
            )


def check_member(member, reporter, check_ignored):
    """Check documentation for a generic member."""
    check_entity(member, reporter)
    if member.is_documented():
        if check_ignored and not member.is_visible():
            reporter.doc_note(
                member,
                "is documented, but is ignored by Doxygen, because its scope is not documented",
            )
        if member.has_inbody_description():
            reporter.doc_note(member, "has in-body comments, which are ignored")


def check_all(tree, reporter, check_ignored):
    """Do all checks for the GROMACS tree."""
    for classobj in tree.get_classes():
        check_class(classobj, reporter)

    for memberobj in tree.get_members():
        check_member(memberobj, reporter, check_ignored)


def main():
    """Run the checking script."""
    parser = OptionParser()
    parser.add_option("-S", "--source-root", help="Source tree root directory")
    parser.add_option("-B", "--build-root", help="Build tree root directory")
    parser.add_option(
        "-l", "--log", help="Write issues into a given log file in addition to stderr"
    )
    parser.add_option("--ignore", help="Set file with patterns for messages to ignore")
    parser.add_option(
        "--check-ignored",
        action="store_true",
        help="Issue notes for comments ignored by Doxygen",
    )
    parser.add_option(
        "-q", "--quiet", action="store_true", help="Do not write status messages"
    )
    parser.add_option(
        "--exitcode",
        action="store_true",
        help="Return non-zero exit code if there are warnings",
    )
    options, args = parser.parse_args()

    reporter = Reporter(options.log)
    if options.ignore:
        reporter.load_filters(options.ignore)

    if not options.quiet:
        sys.stderr.write("Scanning source tree...\n")
    tree = GromacsTree(options.source_root, options.build_root, reporter)
    tree.load_git_attributes()
    if not options.quiet:
        sys.stderr.write("Reading source files...\n")
    # TODO: The checking should be possible without storing everything in memory
    tree.scan_files(keep_contents=True)
    if not options.quiet:
        sys.stderr.write("Finding config.h and other preprocessor macro uses...\n")
    tree.find_define_file_uses()
    if not options.quiet:
        sys.stderr.write("Reading Doxygen XML files...\n")
    tree.load_xml()

    reporter.write_pending()

    if not options.quiet:
        sys.stderr.write("Checking...\n")

    check_all(tree, reporter, options.check_ignored)

    reporter.write_pending()
    reporter.report_unused_filters()
    reporter.close_log()

    if options.exitcode and reporter.had_warnings():
        sys.exit(1)


main()



###########################################################################
#                                                                         #
#    physical_validation,                                                 #
#    a python package to test the physical validity of MD results         #
#                                                                         #
#    Written by Michael R. Shirts <michael.shirts@colorado.edu>           #
#               Pascal T. Merz <pascal.merz@colorado.edu>                 #
#                                                                         #
#    Copyright (C) 2012 University of Virginia                            #
#              (C) 2017 University of Colorado Boulder                    #
#                                                                         #
#    This library is free software; you can redistribute it and/or        #
#    modify it under the terms of the GNU Lesser General Public           #
#    License as published by the Free Software Foundation; either         #
#    version 2.1 of the License, or (at your option) any later version.   #
#                                                                         #
#    This library is distributed in the hope that it will be useful,      #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of       #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    #
#    Lesser General Public License for more details.                      #
#                                                                         #
#    You should have received a copy of the GNU Lesser General Public     #
#    License along with this library; if not, write to the                #
#    Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,     #
#    Boston, MA 02110-1301 USA                                            #
#                                                                         #
###########################################################################
r"""
Data structures carrying simulation data.
"""
import warnings
import numpy as np

import physical_validation.util.error as pv_error


class TrajectoryData(object):
    r"""TrajectoryData: The position and velocity trajectory along the simulation

    The full trajectory is needed to calculate the equipartition of the kinetic energy.
    As they are used in connection, the position and velocity trajectories are expected
    to have the same shape and number of frames.

    The position and velocity trajectories can be accessed either using the getters
    of an object, as in

        * trajectory.position
        * trajectory.velocity

    or using the key notation, as in

        * trajectory['position']
        * trajectory['velocity']

    """

    @staticmethod
    def trajectories():
        return ('position',
                'velocity')

    def __init__(self, position=None, velocity=None):
        self.__position = None
        self.__velocity = None
        self.__nframes = 0

        if position is not None:
            self.position = position
        if velocity is not None:
            self.velocity = velocity

        self.__getters = {
            'position': TrajectoryData.position.__get__,
            'velocity': TrajectoryData.velocity.__get__
        }

        self.__setters = {
            'position': TrajectoryData.position.__set__,
            'velocity': TrajectoryData.velocity.__set__
        }

    def get(self, key):
        return self[key]

    def __getitem__(self, key):
        if key not in self.trajectories():
            raise KeyError
        return self.__getters[key](self)

    def set(self, key, value):
        self[key] = value

    def __setitem__(self, key, value):
        if key not in self.trajectories():
            raise KeyError
        self.__setters[key](self, value)

    @property
    def position(self):
        """Get position"""
        return self.__position

    @position.setter
    def position(self, pos):
        """Set position"""
        pos = np.array(pos)
        if pos.ndim == 2:
            # create 3-dimensional array
            pos = np.array([pos])
        if pos.ndim != 3:
            warnings.warn('Expected 2- or 3-dimensional array.')
        if self.__nframes == 0 and self.__velocity is None:
            self.__nframes = pos.shape[0]
        elif self.__nframes != pos.shape[0]:
            raise pv_error.InputError(['pos'],
                                      'Expected equal number of frames as in velocity trajectory.')
        self.__position = pos

    @property
    def velocity(self):
        """Get velocity"""
        return self.__velocity

    @velocity.setter
    def velocity(self, vel):
        """Set velocity"""
        vel = np.array(vel)
        if vel.ndim == 2:
            # create 3-dimensional array
            vel = np.array([vel])
        if vel.ndim != 3:
            warnings.warn('Expected 2- or 3-dimensional array.')
        if self.__nframes == 0 and self.__position is None:
            self.__nframes = vel.shape[0]
        elif self.__nframes != vel.shape[0]:
            raise pv_error.InputError(['vel'],
                                      'Expected equal number of frames as in position trajectory.')
        self.__velocity = vel

    @property
    def nframes(self):
        """Get number of frames"""
        return self.__nframes



#!/usr/bin/env python3
#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2013- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

"""Checks and/or updates copyright headers in GROMACS source files.

It is used internally by several bash scripts to do copyright-relates tasks,
but can also be invoked directly for some rare use cases.

See docs/dev-manual/code-formatting.rst for more details.
"""

import datetime
import os.path
import re
import sys

from optparse import OptionParser


class CopyrightState(object):
    """Information about an existing (or non-existing) copyright header."""

    def __init__(self, has_copyright, is_correct, is_newstyle, first_year):
        self.has_copyright = has_copyright
        self.is_correct = is_correct
        self.is_newstyle = is_newstyle
        self.first_year = first_year


class CopyrightChecker(object):
    """Logic for analyzing existing copyright headers and generating new ones."""

    _header = ["", "This file is part of the GROMACS molecular simulation package.", ""]
    _copyright = "Copyright {0}- The GROMACS Authors"
    _footer = """
and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.

GROMACS is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public License
as published by the Free Software Foundation; either version 2.1
of the License, or (at your option) any later version.

GROMACS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with GROMACS; if not, see
https://www.gnu.org/licenses, or write to the Free Software Foundation,
Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.

If you want to redistribute modifications to GROMACS, please
consider that scientific software is very special. Version
control is crucial - bugs must be traceable. We will be happy to
consider code for inclusion in the official distribution, but
derived work must not be called official GROMACS. Details are found
in the README & COPYING files - if they are missing, get the
official version at https://www.gromacs.org.

To help us fund GROMACS development, we humbly ask that you cite
the research papers on the package. Check out https://www.gromacs.org.
""".strip().splitlines()

    def check_copyright(self, comment_block):
        """Analyze existing copyright header for correctness and extract information."""
        copyright_re_new = r"Copyright ([0-9]{4})- The GROMACS Authors"
        copyright_re_old = r"Copyright \(c\) (([0-9]{4}[,-])*[0-9]{4}),?.*"
        has_copyright = False
        is_newstyle = True
        is_correct = True
        found_old_style_line = False
        next_header_line = 0
        next_footer_line = 0
        first_year = ""
        for line in comment_block:
            if "Copyright" in line:
                has_copyright = True
                old_match = re.match(copyright_re_old, line)
                match = re.match(copyright_re_new, line)
                if (old_match and not match) and not found_old_style_line:
                    is_newstyle = False
                    found_old_style_line = True
                    get_first_year_re = r"Copyright \(c\) ([0-9]{4}[,-]).*"
                    first_year_line_match = re.match(get_first_year_re, line)
                    first_year_line = first_year_line_match.group(1)
                    # decide which separator to use for finding first year
                    if "-" in first_year_line:
                        pos = first_year_line.find("-")
                        first_year = first_year_line[:pos]
                    elif "," in first_year_line:
                        pos = first_year_line.find(",")
                        first_year = first_year_line[:pos]
                    else:
                        raise Exception("Can't find expected separator!")
                elif match:
                    first_year = match.group(1)
                    new_line = self._copyright.format(first_year)
                    if line != new_line:
                        is_correct = False
                if next_header_line != -1 or next_footer_line != 0:
                    is_correct = False
                continue
            if line.startswith("Written by the Gromacs development team"):
                has_copyright = True
            if next_header_line >= 0:
                if line == self._header[next_header_line]:
                    next_header_line += 1
                    if next_header_line >= len(self._header):
                        next_header_line = -1
                else:
                    is_correct = False
                    is_newstyle = False
            elif next_footer_line >= 0:
                if line == self._footer[next_footer_line]:
                    next_footer_line += 1
                    if next_footer_line >= len(self._footer):
                        next_footer_line = -1
                else:
                    is_correct = False
            else:
                is_correct = False
        if next_header_line != -1 or next_footer_line != -1:
            is_correct = False

        return CopyrightState(has_copyright, is_correct, is_newstyle, first_year)

    def process_copyright(self, state, options, first_year, current_year, reporter):
        """Determine whether a copyright header needs to be updated and report issues."""
        need_update = False

        if state.first_year:
            first_year = state.first_year
            if not current_year >= first_year:
                if options.update_year:
                    need_update = True
                    first_year = current_year
                if options.check or not need_update:
                    reporter.report("initial copyright year wrong")
                else:
                    reporter.report("initial copyright year corrected")

        if not state.has_copyright:
            if options.add_missing:
                need_update = True
            if options.check or not need_update:
                reporter.report("copyright header missing")
            elif options.add_missing:
                reporter.report("copyright header added")
        else:
            if not state.is_newstyle:
                if options.replace_header:
                    need_update = True
                if options.check or not need_update:
                    reporter.report("copyright header incorrect")
                else:
                    reporter.report("copyright header replaced")
            elif not state.is_correct:
                if options.update_header:
                    need_update = True
                if options.check or not need_update:
                    reporter.report("copyright header outdated")
                else:
                    reporter.report("copyright header updated")

        return need_update, first_year

    def get_copyright_text(self, first_year):
        """Construct a new copyright header."""
        output = []
        output.extend(self._header)

        output.append(self._copyright.format(first_year))
        output.extend(self._footer)
        return output


class Reporter(object):
    """Wrapper for reporting issues in a file."""

    def __init__(self, reportfile, filename):
        self._reportfile = reportfile
        self._filename = filename

    def report(self, text):
        self._reportfile.write(self._filename + ": " + text + "\n")


class CommentHandlerC(object):
    """Handler for extracting and creating C-style comments."""

    def extract_first_comment_block(self, content_lines):
        if not content_lines or not content_lines[0].startswith("/*"):
            return ([], 0)
        comment_block = [content_lines[0][2:].strip()]
        line_index = 1
        while line_index < len(content_lines):
            line = content_lines[line_index]
            if "*/" in content_lines[line_index]:
                break
            comment_block.append(line.lstrip("* ").rstrip())
            line_index += 1
        return (comment_block, line_index + 1)

    def create_comment_block(self, lines):
        output = []
        output.append(("/* " + lines[0]).rstrip())
        output.extend([(" * " + x).rstrip() for x in lines[1:]])
        output.append(" */")
        return output


class CommentHandlerSimple(object):
    """Handler for extracting and creating sh-style comments.

    Also other comments of the same type, but with a different comment
    character are supported."""

    def __init__(self, comment_char):
        self._comment_char = comment_char

    def extract_first_comment_block(self, content_lines):
        if not content_lines or not content_lines[0].startswith(self._comment_char):
            return ([], 0)
        comment_block = []
        line_index = 0
        while line_index < len(content_lines):
            line = content_lines[line_index]
            if not line.startswith(self._comment_char):
                break
            comment_block.append(line.lstrip(self._comment_char + " ").rstrip())
            line_index += 1
            if (
                line
                == self._comment_char
                + " the research papers on the package. Check out https://www.gromacs.org."
            ):
                break
        while line_index < len(content_lines):
            line = content_lines[line_index].rstrip()
            if len(line) > 0 and line != self._comment_char:
                break
            line_index += 1
        return (comment_block, line_index)

    def create_comment_block(self, lines):
        output = []
        output.extend([(self._comment_char + " " + x).rstrip() for x in lines])
        output.append("")
        return output


comment_handlers = {
    "c": CommentHandlerC(),
    "tex": CommentHandlerSimple("%"),
    "sh": CommentHandlerSimple("#"),
}


def select_comment_handler(override, filename):
    """Select comment handler for a file based on file name and input options."""
    filetype = override
    if not filetype and filename != "-":
        basename = os.path.basename(filename)
        root, ext = os.path.splitext(basename)
        if ext == ".cmakein":
            dummy, ext2 = os.path.splitext(root)
            if ext2:
                ext = ext2
        if ext in (
            ".c",
            ".cu",
            ".cpp",
            ".cl",
            ".h",
            ".hpp",
            ".cuh",
            ".clh",
            ".y",
            ".l",
            ".pre",
            ".bm",
        ):
            filetype = "c"
        elif ext in (".tex",):
            filetype = "tex"
        elif basename in ("CMakeLists.txt", "GMXRC", "git-pre-commit") or ext in (
            ".cmake",
            ".cmakein",
            ".in",
            ".py",
            ".sh",
            ".bash",
            ".csh",
            ".zsh",
        ):
            filetype = "sh"
    if filetype in comment_handlers:
        return comment_handlers[filetype]
    if filetype:
        sys.stderr.write("Unsupported input format: {0}\n".format(filetype))
    elif filename != "-":
        sys.stderr.write("Unsupported input format: {0}\n".format(filename))
    else:
        sys.stderr.write("No file name or file type provided.\n")
    sys.exit(1)


def create_copyright_header(current_year, language="c"):
    if language not in comment_handlers:
        sys.stderr.write("Unsupported language: {0}\n".format(language))
        sys.exit(1)
    copyright_checker = CopyrightChecker()
    comment_handler = comment_handlers[language]
    copyright_lines = copyright_checker.get_copyright_text(current_year)
    comment_lines = comment_handler.create_comment_block(copyright_lines)
    return "\n".join(comment_lines) + "\n"


def process_options():
    """Process input options."""
    parser = OptionParser()
    parser.add_option("-l", "--lang", help="Comment type to use (c or sh)")
    parser.add_option(
        "--first_year", help="Value for initial year to set in copyright file."
    )
    parser.add_option("-F", "--files", help="File to read list of files from")
    parser.add_option(
        "--check",
        action="store_true",
        help="Do not modify the files, only check the copyright (default action). "
        + "If specified together with --update, do the modifications "
        + "but produce output as if only --check was provided.",
    )
    parser.add_option(
        "--update-year",
        action="store_true",
        help="Update the copyright year if outdated",
    )
    parser.add_option(
        "--update-header",
        action="store_true",
        help="Update the copyright header if outdated",
    )
    parser.add_option(
        "--replace-header",
        action="store_true",
        help="Replace any copyright header with the current one",
    )
    parser.add_option(
        "--remove-old-copyrights",
        action="store_true",
        help="Remove copyright statements not in the new format",
    )
    parser.add_option(
        "--add-missing", action="store_true", help="Add missing copyright headers"
    )
    options, args = parser.parse_args()

    filenames = args
    if options.files:
        with open(options.files, "r") as filelist:
            filenames = [x.strip() for x in filelist.read().splitlines()]
    elif not filenames:
        filenames = ["-"]

    # Default is --check if nothing provided.
    if (
        not options.check
        and not options.update_year
        and not options.update_header
        and not options.replace_header
        and not options.add_missing
    ):
        options.check = True

    return options, filenames


def main():
    """Do processing as a stand-alone script."""
    options, filenames = process_options()
    first_year = options.first_year
    current_year = str(datetime.date.today().year)

    checker = CopyrightChecker()

    # Process each input file in turn.
    for filename in filenames:
        comment_handler = select_comment_handler(options.lang, filename)

        # Read the input file.  We are doing an in-place operation, so can't
        # operate in pass-through mode.
        if filename == "-":
            contents = sys.stdin.read().splitlines()
            reporter = Reporter(sys.stderr, "<stdin>")
        else:
            with open(filename, "r", encoding="utf-8") as inputfile:
                contents = inputfile.read().splitlines()
            reporter = Reporter(sys.stdout, filename)

        output = []
        # Keep lines that must be at the beginning of the file and skip them in
        # the check.
        if contents and (
            contents[0].startswith("#!/")
            or contents[0].startswith("%code requires")
            or contents[0].startswith("/* #if")
        ):
            output.append(contents[0])
            contents = contents[1:]
        # Remove and skip empty lines at the beginning.
        while contents and len(contents[0]) == 0:
            contents = contents[1:]

        # Analyze the first comment block in the file.
        comment_block, line_count = comment_handler.extract_first_comment_block(
            contents
        )
        state = checker.check_copyright(comment_block)
        need_update, first_year = checker.process_copyright(
            state, options, first_year, current_year, reporter
        )
        if options.remove_old_copyrights:
            need_update = True
            reporter.report("old copyrights removed")

        if need_update:
            # Remove the original comment if it was a copyright comment.
            if state.has_copyright:
                contents = contents[line_count:]
            new_block = checker.get_copyright_text(first_year)
            output.extend(comment_handler.create_comment_block(new_block))

        # Write the output file if required.
        if need_update or filename == "-":
            # Append the rest of the input file as it was.
            output.extend(contents)
            output = "\n".join(output) + "\n"
            if filename == "-":
                sys.stdout.write(output)
            else:
                with open(filename, "w", encoding="utf-8") as outputfile:
                    outputfile.write(output)


if __name__ == "__main__":
    main()



#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2019- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

"""Configuration and fixtures for pytest."""

import json
import logging
import os

import pytest

pytest_plugins = ("gmxapi.testsupport",)

try:
    from mpi4py import MPI

    rank_number = MPI.COMM_WORLD.Get_rank()
    comm_size = MPI.COMM_WORLD.Get_size()
except ImportError:
    rank_number = 0
    comm_size = 1
    rank_tag = ""
    MPI = None
else:
    rank_tag = "rank{}:".format(rank_number)

old_factory = logging.getLogRecordFactory()


def record_factory(*args, **kwargs):
    record = old_factory(*args, **kwargs)
    record.rank_tag = rank_tag
    return record


logging.setLogRecordFactory(record_factory)


@pytest.fixture(scope="session")
def spc_water_box_collection(gmxcli, tmp_path_factory):
    """Provide a collection of simulation input items for a simple simulation.

    Prepare the MD input in a freshly created working directory.
    Solvate a 5nm cubic box with spc water. Return a dictionary of the artifacts produced.
    """
    import gmxapi.runtime
    from gmxapi.testsupport import scoped_chdir

    # TODO: (#2896) Fetch MD input from package / library data.
    # Example:
    #     import pkg_resources
    #     # Note: importing pkg_resources means setuptools is required for running this test.
    #     # Get or build TPR file from data bundled via setup(package_data=...)
    #     # Ref https://setuptools.readthedocs.io/en/latest/setuptools.html#including-data-files
    #     from gmx.data import tprfilename

    with scoped_chdir(tmp_path_factory.mktemp("spc_water_box")) as tempdir:
        testdir = os.path.dirname(__file__)
        with open(os.path.join(testdir, "testdata.json"), "r") as fh:
            testdata = json.load(fh)

        # TODO: (#2756) Don't rely on so many automagical behaviors (as described in comments below)

        structurefile = os.path.join(tempdir, "structure.gro")
        # We let `gmx solvate` use the default solvent. Otherwise, we would do
        #     gro_input = testdata['solvent_structure']
        #     with open(structurefile, 'w') as fh:
        #         fh.write('\n'.join(gro_input))
        #         fh.write('\n')

        topfile = os.path.join(tempdir, "topology.top")
        top_input = testdata["solvent_topology"]
        # `gmx solvate` will append a line to the provided file with the molecule count,
        # so we strip the last line from the input topology.
        with open(topfile, "w") as fh:
            fh.write("\n".join(top_input[:-1]))
            fh.write("\n")

        assert os.path.exists(topfile)
        solvate = gmxapi.commandline_operation(
            gmxcli,
            arguments=["solvate", "-box", "5", "5", "5"],
            # We use the default solvent instead of specifying one.
            # input_files={'-cs': structurefile},
            output_files={
                "-p": topfile,
                "-o": structurefile,
            },
        )
        assert os.path.exists(topfile)

        if solvate.output.returncode.result() != 0:
            logging.debug(solvate.output.stderr.result())
            raise RuntimeError("solvate failed in spc_water_box testing fixture.")

        # Choose an exactly representable dt of 2^-9 ps (approximately 0.002)
        dt = 2.0**-9.0
        mdp_input = [
            ("integrator", "md"),
            ("dt", dt),
            ("cutoff-scheme", "Verlet"),
            ("nstcalcenergy", 1),
            ("nsteps", 2),
            ("nstfout", 1),
            ("nstlist", 1),
            ("nstxout", 1),
            ("nstvout", 1),
            ("tcoupl", "v-rescale"),
            ("tc-grps", "System"),
            ("tau-t", 1),
            ("ref-t", 298),
        ]
        mdp_input = "\n".join(
            [" = ".join([str(item) for item in kvpair]) for kvpair in mdp_input]
        )
        mdpfile = os.path.join(tempdir, "md.mdp")
        with open(mdpfile, "w") as fh:
            fh.write(mdp_input)
            fh.write("\n")
        tprfile = os.path.join(tempdir, "topol.tpr")
        # We don't use mdout_mdp, but if we don't specify it to grompp,
        # it will be created in the current working directory.
        mdout_mdp = os.path.join(tempdir, "mdout.mdp")

        grompp = gmxapi.commandline_operation(
            gmxcli,
            "grompp",
            input_files={
                "-f": mdpfile,
                "-p": solvate.output.file["-p"],
                "-c": solvate.output.file["-o"],
                "-po": mdout_mdp,
            },
            output_files={"-o": tprfile},
        )
        if grompp.output.returncode.result() != 0:
            logging.error(grompp.output.stdout.result())
            logging.error(grompp.output.stderr.result())
            raise RuntimeError("grompp failed in spc_water_box testing fixture.")
        logging.debug(grompp.output.stdout.result())

        tprfilename = grompp.output.file["-o"].result()
        assert os.path.exists(tprfilename)
        collection = {
            "tpr_filename": tprfilename,
            "mdp_input_filename": mdpfile,
            "mdp_output_filename": mdout_mdp,
            "topology_filename": solvate.output.file["-p"].result(),
            "gro_filename": solvate.output.file["-o"].result(),
            "mdp_input_list": mdp_input,
        }
        yield collection


@pytest.fixture(scope="session")
def spc_water_box(spc_water_box_collection):
    """Provide a TPR input file for a simple simulation."""
    yield spc_water_box_collection["tpr_filename"]



#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2022- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

"""Provide some additional utilities."""
import functools
import os
from types import MappingProxyType

from gmxapi.operation import function_wrapper


@function_wrapper(output={"path": str})
def join_path(first: str, second: str, output=None):
    """Get a Future path for use in data flow.

    This is useful when a base path or filename is not known until runtime.

    Attributes:
        output.path (str): *first* and *second*, joined by the native filesystem path separator.
    """
    output.path = os.path.join(first, second)


@functools.lru_cache()
def config():
    """Get the GROMACS configuration detected during installation.

    Returns read-only dictionary proxy to file written during installation.
    The :py:class:`~typing.Mapping` contains information about the supporting
    GROMACS installation that was used to configure the Python package
    installation. The exact keys in the Mapping is not formally specified,
    and mostly for internal use.

    .. versionadded:: 0.4

    """
    import json
    from importlib.resources import open_text

    with open_text("gmxapi", "gmxconfig.json") as textfile:
        _config = json.load(textfile)
    return MappingProxyType(_config)



# The myplugin module must be locatable by Python.
# If you configured CMake in the build directory ``/path/to/repo/build`` then,
# assuming you are in ``/path/to/repo``, run the tests with something like
#     PYTHONPATH=./cmake-build-debug/src/pythonmodule mpiexec -n 2 python -m mpi4py -m pytest tests/

# This test is not currently run automatically in any way. Build the module, point your PYTHONPATH at it,
# and run pytest in the tests directory.

import logging
import os

try:
    import mpi4py.MPI as _MPI

    rank_number = _MPI.COMM_WORLD.Get_rank()
    comm_size = _MPI.COMM_WORLD.Get_size()
except (ImportError, ModuleNotFoundError):
    rank_number = 0
    comm_size = 1
    _MPI = None

import gmxapi as gmx
from gmxapi.simulation.context import Context
from gmxapi.simulation.workflow import WorkElement, from_tpr
from gmxapi import version as gmx_version
import pytest

# create console handler
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
# create formatter and add it to the handler
formatter = logging.Formatter("%(asctime)s:%(name)s:%(levelname)s: %(message)s")
ch.setFormatter(formatter)
# add the handlers to the logger
logging.getLogger().addHandler(ch)

logger = logging.getLogger()

mpi_support = pytest.mark.skipif(
    comm_size > 1
    and gmx.utility.config()["gmx_mpi_type"] == "library"
    and not gmx.version.has_feature("mpi_comm_integration"),
    reason="Multi-rank MPI contexts require gmxapi 0.4.",
)


def test_import():
    # Suppress inspection warning outside of testing context.
    # noinspection PyUnresolvedReferences
    import myplugin

    assert myplugin


@mpi_support
@pytest.mark.usefixtures("cleandir")
def test_binding_protocol(spc_water_box, mdrun_kwargs):
    """Test that gmxapi successfully attaches MD plugins."""
    import myplugin

    tpr_filename = spc_water_box
    logger.info(
        "Testing plugin potential with input file {}".format(
            os.path.abspath(tpr_filename)
        )
    )

    assert gmx.version.api_is_at_least(0, 2, 1)
    md = from_tpr([tpr_filename] * comm_size, append_output=False, **mdrun_kwargs)

    potential = WorkElement(
        namespace="myplugin", operation="null_restraint", params={"sites": [1, 4]}
    )
    potential.name = "null restraint"
    md.add_dependency(potential)

    context = Context(md)

    with context as session:
        session.run()

    # See also #3038, #3145, #4079
    assert isinstance(context.potentials, list)
    assert len(context.potentials) > 0
    for restraint in context.potentials:
        if isinstance(restraint, myplugin.NullRestraint):
            assert gmx.version.api_is_at_least(0, 2, 1)
            assert restraint.count() > 1


@mpi_support
@pytest.mark.usefixtures("cleandir")
def test_ensemble_potential_nompi(spc_water_box, mdrun_kwargs):
    """Test ensemble potential without an ensemble."""
    tpr_filename = spc_water_box
    logger.info(
        "Testing plugin potential with input file {}".format(
            os.path.abspath(tpr_filename)
        )
    )

    assert gmx.version.api_is_at_least(0, 0, 5)
    md = from_tpr([tpr_filename], append_output=False, **mdrun_kwargs)

    # Create a WorkElement for the potential
    params = {
        "sites": [1, 4],
        "nbins": 10,
        "binWidth": 0.1,
        "min_dist": 0.0,
        "max_dist": 10.0,
        "experimental": [1.0] * 10,
        "nsamples": 1,
        "sample_period": 0.001,
        "nwindows": 4,
        "k": 10000.0,
        "sigma": 1.0,
    }
    potential = WorkElement(
        namespace="myplugin", operation="ensemble_restraint", params=params
    )
    # Note that we could flexibly capture accessor methods as workflow elements, too. Maybe we can
    # hide the extra Python bindings by letting myplugin.HarmonicRestraint automatically convert
    # to a WorkElement when add_dependency is called on it.
    potential.name = "ensemble_restraint"
    md.add_dependency(potential)

    context = Context(md)

    with context as session:
        session.run()


@mpi_support
@pytest.mark.withmpi_only
@pytest.mark.usefixtures("cleandir")
def test_ensemble_potential_withmpi(spc_water_box, mdrun_kwargs):
    tpr_filename = spc_water_box

    logger.info(
        "Testing plugin potential with input file {}".format(
            os.path.abspath(tpr_filename)
        )
    )

    assert gmx_version.api_is_at_least(0, 0, 5)
    md = from_tpr([tpr_filename, tpr_filename], append_output=False, **mdrun_kwargs)

    # Create a WorkElement for the potential
    params = {
        "sites": [1, 4],
        "nbins": 10,
        "binWidth": 0.1,
        "min_dist": 0.0,
        "max_dist": 10.0,
        "experimental": [0.5] * 10,
        "nsamples": 1,
        "sample_period": 0.001,
        "nwindows": 4,
        "k": 10000.0,
        "sigma": 1.0,
    }

    potential = WorkElement(
        namespace="myplugin", operation="ensemble_restraint", params=params
    )
    # Note that we could flexibly capture accessor methods as workflow elements, too. Maybe we can
    # hide the extra Python bindings by letting myplugin.HarmonicRestraint automatically convert
    # to a WorkElement when add_dependency is called on it.
    potential.name = "ensemble_restraint"
    md.add_dependency(potential)

    context = Context(md)
    with context as session:
        session.run()



#!/usr/bin/env python3
#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2013- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

# This script is used by the GROMACS developers to generate the .cpp
# files that instiantiate all the template specializations of the kernel.
# This script also generates .h files with a list of external template
# declarations and a table of function points. This script is not called
# at CMake time, and users should never need to use it. The generated
# files are versions of the *.pre files in this directory, customized
# for the kernel structure type and/or the detailed kernel type. These
# are:
#
#   A single header file that declares all the kernel function template
#   specializations for this NBNxM kernel layout and a function pointer table.
#
#   Many C++ kernel files, each defining a single kernel function template
#   specialization. These functions can take a noticeable time to compile,
#   and should be in separate files to take advantage of make-time parallelism.
#
# These files are written to two separate directories for the two kernel
# layouts 2xMM and 4xM.
#
# Note that while functions for both NBNxM kernel layouts are compiled
# and built into an mdrun executable, because that executable
# is not portable, only the functions for the useful NBNxM kernel
# structure for the hardware selected at CMake time contain real
# kernel logic. A run-time error occurs if an inappropriate kernel
# dispatcher function is called (but that is normally impossible).

import re
import sys
import os

os.chdir(os.path.dirname(os.path.abspath(__file__)))
import collections  # Requires Python 2.7

sys.path.append("../../../../admin")
from copyright import create_copyright_header

FileHeader = create_copyright_header("2012")
FileHeader += """/*
 * Note: this file was generated by the Verlet kernel generator for
 * kernel type {0}.
 */

"""


def read_kernel_template(filename):
    with open(filename, "r") as TemplateFile:
        TemplateText = TemplateFile.read()
    copyright_re = r"/\*\n \* This file is part of the GROMACS molecular simulation package\.\n( \*.*\n)* \*/\n"
    match = re.match(copyright_re, TemplateText)
    if match:
        TemplateText = TemplateText[match.end() :]
    return TemplateText


# The dict order must match the order of an enumeration in
# kernel_simd_template.c.pre
ElectrostaticsDict = collections.OrderedDict()
ElectrostaticsDict["ElecRF"] = {"param": "KernelCoulombType::RF, VdwCutoffCheck::No"}
ElectrostaticsDict["ElecQSTab"] = {
    "param": "KernelCoulombType::EwaldTabulated, VdwCutoffCheck::No"
}
ElectrostaticsDict["ElecQSTabTwinCut"] = {
    "param": "KernelCoulombType::EwaldTabulated, VdwCutoffCheck::Yes"
}
ElectrostaticsDict["ElecEw"] = {
    "param": "KernelCoulombType::EwaldAnalytical, VdwCutoffCheck::No"
}
ElectrostaticsDict["ElecEwTwinCut"] = {
    "param": "KernelCoulombType::EwaldAnalytical, VdwCutoffCheck::Yes"
}
ElectrostaticsDict["ElecNone"] = {
    "param": "KernelCoulombType::None, VdwCutoffCheck::Yes"
}


# The dict order must match the order of a C enumeration.
VdwTreatmentDict = collections.OrderedDict()
VdwTreatmentDict["VdwLJCombGeom"] = {
    "param": "LJCombinationRule::Geometric, InteractionModifiers::PotShift, LJEwald::None"
}
VdwTreatmentDict["VdwLJCombLB"] = {
    "param": "LJCombinationRule::LorentzBerthelot, InteractionModifiers::PotShift, LJEwald::None"
}
VdwTreatmentDict["VdwLJ"] = {
    "param": "LJCombinationRule::None, InteractionModifiers::PotShift, LJEwald::None"
}
VdwTreatmentDict["VdwLJFSw"] = {
    "param": "LJCombinationRule::None, InteractionModifiers::ForceSwitch, LJEwald::None"
}
VdwTreatmentDict["VdwLJPSw"] = {
    "param": "LJCombinationRule::None, InteractionModifiers::PotSwitch, LJEwald::None"
}
VdwTreatmentDict["VdwLJEwCombGeom"] = {
    "param": "LJCombinationRule::None, InteractionModifiers::PotShift, LJEwald::CombGeometric"
}

# This is OK as an unordered dict
EnergiesComputationDict = {
    "F": {
        "function type": "void",
        "param": "EnergyOutput::None",
    },
    "VF": {
        "function type": "void",
        "param": "EnergyOutput::System",
    },
    "VgrpF": {
        "function type": "void",
        "param": "EnergyOutput::GroupPairs",
    },
}

# This is OK as an unordered dict
VerletKernelTypeDict = {
    "2xmm": {
        "param": "KernelLayout::r2xMM",
        "define": "GMX_HAVE_NBNXM_SIMD_2XMM",
    },
    "4xm": {
        "param": "KernelLayout::r4xM",
        "define": "GMX_HAVE_NBNXM_SIMD_4XM",
    },
}

KernelsHeaderTemplate = read_kernel_template("kernel_simd_template.h.pre")

# For each Verlet kernel type, write two kinds of files:
#   a header file defining the functions for all the kernels and
#     the kernel function lookup table
#   for each kernel, a file defining the single C function for that kernel
for type in VerletKernelTypeDict:
    DirName = "../kernels_simd_{0}".format(type)
    KernelNamePrefix = "nbnxmKernelSimd"
    KernelFileNamePrefix = "nbnxm_kernel"
    KernelFileNamePrefix = "kernel"
    KernelsFileName = "{0}_simd_{1}".format(KernelFileNamePrefix, type)
    KernelsHeaderFileName = "kernels.h"
    KernelsHeaderPathName = "{1}".format(type, KernelsHeaderFileName)
    KernelFunctionLookupTable = {}
    KernelDeclarations = ""
    KernelTemplate = read_kernel_template("kernel_simd_template.cpp.pre")

    # Loop over all kernels
    for ener in EnergiesComputationDict:
        KernelFunctionLookupTable[ener] = ""
        for elec in ElectrostaticsDict:
            if elec == "ElecNone":
                KernelFunctionLookupTable[ener] += "#    if GMX_USE_EXT_FMM\n"
            KernelFunctionLookupTable[ener] += "    {\n"
            for ljtreat in VdwTreatmentDict:
                KernelName = "{0}<{1}, {2}, {3}, {4}>".format(
                    KernelNamePrefix,
                    VerletKernelTypeDict[type]["param"],
                    ElectrostaticsDict[elec]["param"],
                    VdwTreatmentDict[ljtreat]["param"],
                    EnergiesComputationDict[ener]["param"],
                )
                KernelFileName = "{0}_{1}_{2}_{3}".format(
                    KernelFileNamePrefix, elec, ljtreat, ener, type
                )

                # Declare the kernel function
                KernelDeclarations += "extern template {1}\n        {0}(\n".format(
                    KernelName, EnergiesComputationDict[ener]["function type"]
                )
                KernelDeclarations += "        const NbnxnPairlistCpu&    pairlist,\n        const nbnxn_atomdata_t&    nbat,\n        const interaction_const_t& ic,\n        const rvec*                shift_vec,\n        nbnxn_atomdata_output_t*   out);\n\n"

                # Write the file with the kernel definition
                with open(
                    "{0}/{1}.cpp".format(DirName, KernelFileName), "w"
                ) as kernelfp:
                    kernelfp.write(FileHeader.format(type))
                    kernelfp.write(
                        KernelTemplate.format(
                            VerletKernelTypeDict[type]["param"],
                            ElectrostaticsDict[elec]["param"],
                            VdwTreatmentDict[ljtreat]["param"],
                            EnergiesComputationDict[ener]["param"],
                            VerletKernelTypeDict[type]["define"],
                        )
                    )

                # Enter the kernel function in the lookup table
                KernelFunctionLookupTable[ener] += "            {0},\n".format(
                    KernelName
                )

            KernelFunctionLookupTable[ener] += "    },\n"
            if elec == "ElecNone":
                KernelFunctionLookupTable[ener] += "#    endif\n"

        KernelDeclarations += "\n"

    # Write the header file that declares all the kernel
    # functions for this type
    with open("{0}/{1}".format(DirName, KernelsHeaderFileName), "w") as fp:
        fp.write(FileHeader.format(type))
        fp.write(
            KernelsHeaderTemplate.format(
                KernelDeclarations,
                type,
                KernelFunctionLookupTable["F"],
                KernelFunctionLookupTable["VF"],
                KernelFunctionLookupTable["VgrpF"],
            )
        )

sys.exit()



#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2019- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

"""gmxapi data types and interfaces.

TBD: https://gitlab.com/gromacs/gromacs/-/issues/2993
"""

__all__ = ["ndarray", "NDArray"]

import collections.abc
import typing

import gmxapi.abc
from gmxapi import exceptions


_T = typing.TypeVar("_T")


class NDArray(gmxapi.abc.NDArray, typing.Generic[_T]):
    """N-Dimensional array type."""

    dtype: typing.Type[_T] = None
    shape: typing.Tuple[int] = ()

    def __init__(self, data: typing.Union[_T, typing.Sequence[_T]] = None):
        if data is None:
            self._values = []
            self.dtype = None
            self.shape = (0,)
        else:
            if hasattr(data, "result") or (
                isinstance(data, collections.abc.Iterable)
                and any([hasattr(item, "result") for item in data])
            ):
                raise exceptions.ValueError(
                    "Make a Future of type NDArray instead of NDArray of type Future, or call result() first."
                )
            if isinstance(data, (str, bytes)):
                data = [data]
                length = 1
            else:
                try:
                    length = len(data)
                except TypeError:
                    # data is a scalar
                    length = 1
                    data = [data]
            self._values = data
            if length > 0:
                self.dtype: typing.Type[_T] = type(data[0])
                self.shape = (length,)
        if len(self._values) > 0 or len(self.shape) > 0:
            assert self.shape[0] == len(self._values)

    def to_list(self):
        return self._values

    def __repr__(self):
        return f"<gmxapi.NDArray: dtype={self.dtype}, shape={self.shape}>"

    def __getitem__(self, i: int) -> _T:
        return self._values[i]

    def __len__(self) -> int:
        return len(self._values)


class ArrayFuture(gmxapi.abc.Future, typing.Generic[_T]):
    """Annotation type for gmxapi array results.

    gmxapi interfaces for structured data need to be updated. Until array data
    has a more concrete scheme, it is useful to have an abstract class for use
    in static type hints to distinguish Futures for arrays of certain element
    types from Futures of scalars of certain types.
    """

    _dtype: typing.Type[_T]

    @property
    def dtype(self) -> typing.Type[_T]:
        return self._dtype

    def result(self) -> NDArray[_T]:
        raise NotImplementedError


def ndarray(data=None, shape=None, dtype=None):
    """Create an NDArray object from the provided iterable.

    Arguments:
        data: object supporting sequence, buffer, or Array Interface protocol

    ..  versionadded:: 0.1
        *shape* and *dtype* parameters

    If ``data`` is provided, ``shape`` and ``dtype`` are optional. If ``data`` is not
    provided, both ``shape`` and ``dtype`` are required.

    If ``data`` is provided and shape is provided, ``data`` must be compatible with
    or convertible to ``shape``. See Broadcast Rules in `datamodel` documentation.

    If ``data`` is provided and ``dtype`` is not provided, data type is inferred
    as the narrowest scalar type necessary to hold any element in ``data``.
    ``dtype``, whether inferred or explicit, must be compatible with all elements
    of ``data``.

    The returned object implements the gmxapi N-dimensional Array Interface.
    """
    if data is None:
        array = NDArray()
    else:
        if isinstance(data, NDArray):
            return data
        # data is not None, but may still be an empty sequence.
        length = 0
        try:
            length = len(data)
        except TypeError:
            # data is a scalar
            length = 1
            data = [data]
        array = NDArray(data)
    return array



#!/usr/bin/env python3
#
# This file is part of the GROMACS molecular simulation package.
#
# Copyright 2012- The GROMACS Authors
# and the project initiators Erik Lindahl, Berk Hess and David van der Spoel.
# Consult the AUTHORS/COPYING files and https://www.gromacs.org for details.
#
# GROMACS is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# GROMACS is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with GROMACS; if not, see
# https://www.gnu.org/licenses, or write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
#
# If you want to redistribute modifications to GROMACS, please
# consider that scientific software is very special. Version
# control is crucial - bugs must be traceable. We will be happy to
# consider code for inclusion in the official distribution, but
# derived work must not be called official GROMACS. Details are found
# in the README & COPYING files - if they are missing, get the
# official version at https://www.gromacs.org.
#
# To help us fund GROMACS development, we humbly ask that you cite
# the research papers on the package. Check out https://www.gromacs.org.

"""Generate include dependency graphs.

This script generates include dependency graphs from the GROMACS source tree.
One graph is generated to show inter-module dependencies, and separate graphs
for each module to show file-level dependencies within the module.

Output format for the graphs is suitable for processing with 'dot' in graphviz.

The graphs are built from the source tree representation constructed in
gmxtree.py.

Classes Graph, Node, Edge, and EdgeType provide a relatively general
implementation for constructing 'dot' graphs.  GraphBuilder is used to
create Graph instances from a gmxtree.GromacsTree object; the actual graph
objects will not contain any references to the gmxtree objects.

When run in script mode, the GromacsTree object is first constructed, and then
GraphBuilder is used to construct the necessary graphs, which are then written
out.

The produced graphs are documented in doxygen.md.
"""

import os.path
import re
import functools

from gmxtree import DocType


@functools.total_ordering
class EdgeType(object):
    """Enumeration type for edge types in include dependency graphs."""

    # Mapping to string representation for the internal integer values
    _names = [
        "test",
        "pubimpl",
        "libimpl",
        "library",
        "public",
        "intramodule",
        "legacy",
        "undocumented",
    ]

    def __init__(self, value):
        """Initialize a EdgeType instance.

        EdgeType.{test,pubimpl,...,undocumented} should be used outside the
        class instead of calling the constructor.
        """
        self._value = value

    def __str__(self):
        """Return string representation for the edge type (for debugging)."""
        return self._names[self._value]

    def __eq__(self, other):
        """Order edge types in the order of increasing coupling."""
        return self._value == other._value

    def __lt__(self, other):
        """Order edge types in the order of increasing coupling."""
        return self._value < other._value


# Tests depend on test
EdgeType.test = EdgeType(0)
# Implementation depends on public/library headers
EdgeType.pubimpl = EdgeType(1)
EdgeType.libimpl = EdgeType(2)
# Library header depends on other module
EdgeType.library = EdgeType(3)
# Public header depends on other module
EdgeType.public = EdgeType(4)
# Intramodule dependency
EdgeType.intramodule = EdgeType(5)
EdgeType.legacy = EdgeType(6)
EdgeType.cyclic = EdgeType(7)
# Invalid dependency
EdgeType.undocumented = EdgeType(8)


class Edge(object):
    """Graph edge between two Node objects in 'dot' graph.

    Signifies an include dependency between the two nodes, and manages types
    associated with the dependencies.
    """

    def __init__(self, fromnode, tonode, edgetype):
        """Create edge between given Nodes with given type."""
        self._fromnode = fromnode
        self._tonode = tonode
        self._edgetype = edgetype

    def merge_edge(self, other):
        """Merge another edge into this one and choose an appropriate type.

        Updates the type of this edge based on the types of the merged edges.
        """
        self._edgetype = max(self._edgetype, other._edgetype)

    def format(self):
        """Format this edge for 'dot'."""
        # If you change these styles, update also the legend in modulegraph.md
        if self._fromnode.is_file_node() and self._tonode.is_file_node():
            properties = ""
        elif self._edgetype == EdgeType.intramodule:
            properties = ""
        elif self._edgetype == EdgeType.test:
            # TODO: Consider if only some test edges should be made non-constraints
            properties = 'color=".33 .8 .8", style=dashed, constraint=no'
        elif self._edgetype == EdgeType.libimpl:
            properties = 'color=".66 .8 .8", style=dashed'
        elif self._edgetype == EdgeType.pubimpl:
            properties = "color=black, style=dashed"
        elif self._edgetype == EdgeType.library:
            properties = 'color=".66 .8 .8"'
        elif self._edgetype == EdgeType.public:
            properties = "color=black"
        elif self._edgetype == EdgeType.legacy:
            properties = "color=grey75"
        elif self._edgetype == EdgeType.cyclic:
            properties = "color=red, constraint=no"
        else:  # undocumented
            properties = "color=red"
        return "{0} -> {1} [{2}]".format(
            self._fromnode.get_nodename(), self._tonode.get_nodename(), properties
        )


class Node(object):
    """Node in 'dot' graph."""

    def __init__(self, nodename, label, style=None, properties=None, is_file=False):
        """Create node with given attributes.

        is_file does not affect the appearance of the node, but is used for
        formatting edges between two files differently from other edges.
        style and properties should be iterables with graphviz attributes for
        the node.

        Node can have child nodes.  Such nodes are rendered as cluster
        subgraphs for 'dot'.
        """
        self._nodename = nodename
        self._label = label
        if style:
            self._style = ",".join(style)
        else:
            self._style = None
        if properties:
            self._properties = ", ".join(properties)
        else:
            self._properties = None
        self._is_file = is_file
        self._children = []

    def add_child(self, child):
        """Add a child node."""
        self._children.append(child)

    def clear_children(self):
        """Remove all children from the node."""
        self._children = []

    def is_file_node(self):
        """Return True if the node was created with is_file=True."""
        return self._is_file

    def get_nodename(self):
        """Get internal name of the node in 'dot'."""
        return self._nodename

    def get_children(self, recursive=False):
        """Get list of child nodes."""
        if recursive:
            result = list(self._children)
            for child in self._children:
                result.extend(child.get_children(recursive=True))
            return result
        else:
            return self._children

    def format(self):
        """Format this node for 'dot'."""
        # TODO: Take indent as a parameter to make output marginally nicer.
        result = ""
        if self._children:
            result += "    subgraph cluster_{0} {{\n".format(self._nodename)
            result += '        label = "{0}"\n'.format(self._label)
            for child in self._children:
                result += child.format()
            result += "    }\n"
        else:
            properties = 'label="{0}"'.format(self._label)
            if self._properties:
                properties += ", " + self._properties
            if self._style:
                properties += ', style="{0}"'.format(self._style)
            result += "    {0} [{1}]\n".format(self._nodename, properties)
        return result


class Graph(object):
    """Graph for 'dot'."""

    def __init__(self, nodes, edges):
        """Create graph with given nodes and edges."""
        self._nodes = set(nodes)
        self._edges = edges
        self._left_to_right = False
        self._concentrate = True

    def set_options(self, left_to_right=None, concentrate=None):
        """Set output options for the graph."""
        if left_to_right != None:
            self._left_to_right = left_to_right
        if concentrate != None:
            self._concentrate = concentrate

    def merge_nodes(self, nodes, target):
        """Merge a set of nodes into a single node.

        All nodes from the list nodes are merged into the target node.
        All edges to or from the merged nodes are rerouted to/from target
        instead.  Duplicate edges are not created.  Instead, if an edge already
        exists, the edge types are merged.  All nodes from the list nodes are
        removed from the graph after the merge is done.
        """
        nodes = set(nodes)
        nodes.add(target)
        newedges = []
        edgesto = dict()
        edgesfrom = dict()
        for edge in self._edges:
            isfrom = edge._fromnode in nodes
            isto = edge._tonode in nodes
            if isfrom and isto:
                pass
            elif isfrom:
                if not edge._tonode in edgesfrom:
                    edgesfrom[edge._tonode] = Edge(target, edge._tonode, edge._edgetype)
                else:
                    edgesfrom[edge._tonode].merge_edge(edge)
            elif isto:
                if not edge._fromnode in edgesto:
                    edgesto[edge._fromnode] = Edge(
                        edge._fromnode, target, edge._edgetype
                    )
                else:
                    edgesto[edge._fromnode].merge_edge(edge)
            else:
                newedges.append(edge)
        newedges.extend(edgesfrom.values())
        newedges.extend(edgesto.values())
        self._edges = newedges

    def collapse_node(self, node):
        """Merge all children of a node into the node.

        All child nodes are removed after the merge is done.
        """
        nodes = node.get_children(recursive=True)
        self.merge_nodes(nodes, node)
        node.clear_children()

    def write(self, outfile):
        """Write the graph in 'dot' format."""
        outfile.write("digraph includedeps {\n")
        if self._left_to_right:
            outfile.write("    rankdir = LR\n")
        if self._concentrate:
            outfile.write("    concentrate = true\n")
        outfile.write(
            '    node [fontname="FreeSans",fontsize=10,height=.2,' "shape=box]\n"
        )
        for node in self._nodes:
            outfile.write(node.format())
        for edge in self._edges:
            outfile.write("    " + edge.format() + "\n")
        outfile.write("}\n")


class GraphBuilder(object):
    """Builder for Graph objects from gmxtree.GromacsTree representation."""

    def __init__(self, tree):
        """Initialize builder for a given tree representation."""
        self._tree = tree

    def _create_file_node(self, fileobj, filenodes):
        """Create graph node for a file object.

        filenodes is a dict() that maps file objects to their nodes, and is
        updated by this call.
        """
        nodename = re.subn(r"[-./]", "_", fileobj.get_relpath())[0]
        style = []
        properties = []
        properties.append('URL="\\ref {0}"'.format(fileobj.get_name()))
        if not fileobj.get_module():
            style.append("bold")
            properties.append("color=red")
        if fileobj.is_test_file():
            style.append("filled")
            properties.append('fillcolor=".33 .2 1"')
        elif fileobj.is_source_file():
            style.append("filled")
            properties.append("fillcolor=grey75")
        elif fileobj.get_api_type() == DocType.public:
            style.append("filled")
            properties.append('fillcolor=".66 .2 1"')
        elif fileobj.get_api_type() == DocType.library:
            style.append("filled")
            properties.append('fillcolor=".66 .5 1"')
        node = Node(nodename, fileobj.get_name(), style, properties, is_file=True)
        filenodes[fileobj] = node
        return node

    def _get_file_edge_type(self, fromfile, tofile):
        """Get EdgeType for an edge between two file objects.

        Determines the type for the edge from the information provided by
        gmxtree.
        """
        intramodule = fromfile.get_module() == tofile.get_module()
        is_legacy = not tofile.api_type_is_reliable()
        if fromfile.get_module() == tofile.get_module():
            return EdgeType.intramodule
        elif tofile.get_api_type() == DocType.internal and not tofile.is_public():
            if is_legacy:
                return EdgeType.legacy
            else:
                return EdgeType.undocumented
        elif fromfile.is_test_file():
            return EdgeType.test
        elif tofile.is_test_file():
            return EdgeType.undocumented
        elif fromfile.is_module_internal():
            if tofile.is_public():
                return EdgeType.pubimpl
            elif tofile.get_api_type() == DocType.library:
                return EdgeType.libimpl
            elif is_legacy:
                return EdgeType.legacy
            elif not tofile.is_documented():
                return EdgeType.undocumented
            else:
                raise ValueError(
                    "Unknown edge type between {0} and {1}".format(
                        fromfile.get_relpath(), tofile.get_relpath()
                    )
                )
        elif fromfile.get_api_type() == DocType.library:
            return EdgeType.library
        elif fromfile.is_public():
            if tofile.is_public():
                return EdgeType.public
            else:
                return EdgeType.undocumented
        elif is_legacy:
            return EdgeType.legacy
        else:
            raise ValueError(
                "Unknown edge type between {0} and {1}".format(
                    fromfile.get_relpath(), tofile.get_relpath()
                )
            )

    def _create_file_edge(self, fromfile, tofile, filenodes):
        """Create edge between two file objects.

        Determines the type for the edge from the information provided by
        gmxtree.
        """
        edgetype = self._get_file_edge_type(fromfile, tofile)
        return Edge(filenodes[fromfile], filenodes[tofile], edgetype)

    def _create_file_edges(self, filenodes):
        """Create edges between all file nodes.

        Create edges between file nodes specified in filenodes from all include
        dependencies.  An edge is created only if both ends of the dependency
        are in the list of nodes.
        """
        edges = []
        for fileobj in filenodes.keys():
            for includedfile in fileobj.get_includes():
                otherfile = includedfile.get_file()
                if otherfile and otherfile in filenodes:
                    edge = self._create_file_edge(fileobj, otherfile, filenodes)
                    edges.append(edge)
        return edges

    def _get_module_color(self, modulegroup):
        # If you change these styles, update also the legend in modulegraph.md
        if modulegroup == "legacy":
            return "fillcolor=grey75"
        elif modulegroup == "analysismodules":
            return 'fillcolor="0 .2 1"'
        elif modulegroup == "utilitymodules":
            return 'fillcolor=".08 .2 1"'
        elif modulegroup == "mdrun":
            return 'fillcolor=".75 .2 1"'
        return None

    def _create_module_node(self, module):
        """Create node for a module."""
        style = []
        properties = []
        properties.append("shape=ellipse")
        if module.is_documented():
            properties.append('URL="\\ref {0}"'.format(module.get_name()))
        if not module.is_documented():
            fillcolor = self._get_module_color("legacy")
        else:
            fillcolor = self._get_module_color(module.get_group())
        if fillcolor:
            style.append("filled")
            properties.append(fillcolor)
        rootdir = module.get_root_dir()
        nodename = "module_" + re.subn(r"[-./]", "_", rootdir.get_relpath())[0]
        label = module.get_name()[7:]
        node = Node(nodename, label, style, properties)
        return node

    def _create_module_edges(self, modulenodes):
        """Create edges between all module nodes.

        Create edges between module nodes specified in modulenodes from all
        include dependencies.  An edge is created only if both ends of the
        dependency are in the list of nodes.
        """
        edges = []
        for moduleobj in modulenodes.keys():
            for dep in moduleobj.get_dependencies():
                othermodule = dep.get_other_module()
                if othermodule and othermodule in modulenodes:
                    if dep.is_cycle_suppressed():
                        edgetype = EdgeType.cyclic
                    else:
                        edgetype = max(
                            [
                                self._get_file_edge_type(
                                    x.get_including_file(), x.get_file()
                                )
                                for x in dep.get_included_files()
                            ]
                        )
                    edge = Edge(
                        modulenodes[moduleobj], modulenodes[othermodule], edgetype
                    )
                    edges.append(edge)
        return edges

    def create_modules_graph(self):
        """Create module dependency graph."""
        nodes = []
        modulenodes = dict()
        libgromacsnode = Node("libgromacs", "libgromacs")
        nodes.append(libgromacsnode)
        for moduleobj in self._tree.get_modules():
            node = self._create_module_node(moduleobj)
            if moduleobj.get_root_dir().get_relpath().startswith("src/gromacs"):
                libgromacsnode.add_child(node)
            else:
                nodes.append(node)
            modulenodes[moduleobj] = node
        edges = self._create_module_edges(modulenodes)
        graph = Graph(nodes, edges)
        graph.set_options(concentrate=False)
        return graph

    def create_module_file_graph(self, module):
        """Create file dependency graph for files within a module."""
        filenodes = dict()
        nodes = []
        for fileobj in module.get_files():
            nodes.append(self._create_file_node(fileobj, filenodes))
        edges = self._create_file_edges(filenodes)
        graph = Graph(nodes, edges)
        graph.set_options(left_to_right=True)
        return graph


def main():
    """Run the graph generation script."""
    import os
    import sys

    from optparse import OptionParser

    from gmxtree import GromacsTree
    from reporter import Reporter

    parser = OptionParser()
    parser.add_option("-S", "--source-root", help="Source tree root directory")
    parser.add_option("-B", "--build-root", help="Build tree root directory")
    parser.add_option(
        "--ignore-cycles", help="Set file with module dependencies to ignore in cycles"
    )
    parser.add_option(
        "-o", "--outdir", default=".", help="Specify output directory for graphs"
    )
    parser.add_option(
        "-q", "--quiet", action="store_true", help="Do not write status messages"
    )
    options, args = parser.parse_args()

    reporter = Reporter(quiet=True)

    if not options.quiet:
        sys.stderr.write("Scanning source tree...\n")
    tree = GromacsTree(options.source_root, options.build_root, reporter)
    if not options.quiet:
        sys.stderr.write("Reading source files...\n")
    tree.scan_files()
    if options.ignore_cycles:
        tree.load_cycle_suppression_list(options.ignore_cycles)
    if not options.quiet:
        sys.stderr.write("Reading Doxygen XML files...\n")
    tree.load_xml(only_files=True)

    if not options.quiet:
        sys.stderr.write("Writing graphs...\n")
    graphbuilder = GraphBuilder(tree)
    if not os.path.exists(options.outdir):
        os.mkdir(options.outdir)

    filename = os.path.join(options.outdir, "module-deps.dot")
    graph = graphbuilder.create_modules_graph()
    with open(filename, "w") as outfile:
        graph.write(outfile)

    # Skip some modules that are too big to make any sense
    skippedmodules = ("gmxlib", "mdlib", "gmxana", "gmxpreprocess")
    for module in tree.get_modules():
        if not module.get_name()[7:] in skippedmodules:
            filename = "{0}-deps.dot".format(module.get_name())
            filename = os.path.join(options.outdir, filename)
            graph = graphbuilder.create_module_file_graph(module)
            with open(filename, "w") as outfile:
                graph.write(outfile)


if __name__ == "__main__":
    main()



###########################################################################
#                                                                         #
#    physical_validation,                                                 #
#    a python package to test the physical validity of MD results         #
#                                                                         #
#    Written by Michael R. Shirts <michael.shirts@colorado.edu>           #
#               Pascal T. Merz <pascal.merz@colorado.edu>                 #
#                                                                         #
#    Copyright (C) 2012 University of Virginia                            #
#              (C) 2017 University of Colorado Boulder                    #
#                                                                         #
#    This library is free software; you can redistribute it and/or        #
#    modify it under the terms of the GNU Lesser General Public           #
#    License as published by the Free Software Foundation; either         #
#    version 2.1 of the License, or (at your option) any later version.   #
#                                                                         #
#    This library is distributed in the hope that it will be useful,      #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of       #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    #
#    Lesser General Public License for more details.                      #
#                                                                         #
#    You should have received a copy of the GNU Lesser General Public     #
#    License along with this library; if not, write to the                #
#    Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,     #
#    Boston, MA 02110-1301 USA                                            #
#                                                                         #
###########################################################################
"""
The `integratorconvergence` module is part of the physical_validation
package, and consists of checks of the convergence of the MD integrator.
"""
# py2 compatibility
from __future__ import absolute_import, division, print_function

from .util import integrator as util_integ
from .data import SimulationData
from .util import error as pv_error


def convergence(simulations,
                convergence_test='max_deviation',
                verbose=True, slope=False,
                screen=False, filename=None):
    r"""
    Compares the convergence of the fluctuations of conserved quantities
    with decreasing simulation time step to theoretical expectations.

    Parameters
    ----------
    simulations : list of SimulationData
        The (otherwise identical) simulations performed using different
        timesteps
    convergence_test : str, optional
        A function defining the convergence test. Currently, only one
        test is implemented:
        `max_deviation`
    verbose : bool, optional
    screen : bool
        Plot convergence on screen. Default: False.
    filename : string
        Plot convergence to `filename`.pdf. Default: None.

    Returns
    -------
    result : float

    Notes
    -----
    For a simplectic integration algorithm, the fluctuations
    :math:`\delta E` of a constant of motion :math:`E` (such as, for
    example, the total energy in a NVE simulations) are theoretically
    expected to scale like the squared timestep of the integration.
    When comparing two otherwise identical simulations performed at
    different time step :math:`\Delta t`, the following equality is
    hence expected to hold:

    .. math::
        \frac{\Delta t_1}{\Delta t_2} = \frac{\delta E_1}{\delta E_2}

    This function calculates the ratio of the fluctuation for simulations
    performed at different timesteps and compares it to the analytically
    expected value. If the deviation is larger than `tol`, the test is
    considered failed.

    """
    constant_of_motion = {}

    convergence_tests = {
        'max_deviation': util_integ.max_deviation
    }

    if convergence_test not in convergence_tests:
        raise pv_error.InputError('convergence_test',
                                  'Unknown convergence test.')

    convergence_test = convergence_tests[convergence_test]

    for s in simulations:
        if not isinstance(s, SimulationData):
            raise pv_error.InputError('simulations',
                                      'Expected a list of objects of type SimulationData')
        if s.dt <= 0:
            raise pv_error.InputError('simulations',
                                      'Found SimulationData with timestep dt<=0')
        key = str(s.dt)

        if key in constant_of_motion:
            raise pv_error.InputError('simulations',
                                      'Found two simulations with identical timestep')

        constant_of_motion[key] = s.observables.constant_of_motion

    return util_integ.check_convergence(constant_of_motion,
                                        convergence_test=convergence_test,
                                        verbose=verbose, slope=slope,
                                        screen=screen, filename=filename)



###########################################################################
#                                                                         #
#    physical_validation,                                                 #
#    a python package to test the physical validity of MD results         #
#                                                                         #
#    Written by Michael R. Shirts <michael.shirts@colorado.edu>           #
#               Pascal T. Merz <pascal.merz@colorado.edu>                 #
#                                                                         #
#    Copyright (C) 2012 University of Virginia                            #
#              (C) 2017 University of Colorado Boulder                    #
#                                                                         #
#    This library is free software; you can redistribute it and/or        #
#    modify it under the terms of the GNU Lesser General Public           #
#    License as published by the Free Software Foundation; either         #
#    version 2.1 of the License, or (at your option) any later version.   #
#                                                                         #
#    This library is distributed in the hope that it will be useful,      #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of       #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    #
#    Lesser General Public License for more details.                      #
#                                                                         #
#    You should have received a copy of the GNU Lesser General Public     #
#    License along with this library; if not, write to the                #
#    Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,     #
#    Boston, MA 02110-1301 USA                                            #
#                                                                         #
###########################################################################
r"""
Data structures carrying simulation data.
"""
import warnings
import numpy as np

import physical_validation.util.error as pv_error


class ObservableData(object):
    r"""ObservableData: The trajectory of (macroscopic) observables during the simulation

    Stores a number of different observables:
    'kinetic_energy': the kinetic energy of the system,
    'potential_energy': the potential energy of the system,
    'total_energy': the total energy of the system,
    'volume': the volume of the system box,
    'pressure': the pressure of the system,
    'temperature': the temperature of the system,
    'constant_of_motion': a constant of motion of the trajectory.

    The observable trajectories can be accessed either using the getters of an object, as in
        observables.kinetic_energy
    or using the key notation, as in
        observables['kinetic_energy']
    """
    @staticmethod
    def observables():
        return ['kinetic_energy',
                'potential_energy',
                'total_energy',
                'volume',
                'pressure',
                'temperature',
                'constant_of_motion']

    def __init__(self,
                 kinetic_energy=None, potential_energy=None, total_energy=None,
                 volume=None, pressure=None, temperature=None, constant_of_motion=None):
        self.__kinetic_energy = None
        self.__potential_energy = None
        self.__total_energy = None
        self.__volume = None
        self.__pressure = None
        self.__temperature = None
        self.__constant_of_motion = None
        self.__nframes = -1
        self.__kinetic_energy_per_molec = None

        self.kinetic_energy = kinetic_energy
        self.potential_energy = potential_energy
        self.total_energy = total_energy
        self.volume = volume
        self.pressure = pressure
        self.temperature = temperature
        self.constant_of_motion = constant_of_motion

        self.__getters = {
            'kinetic_energy': ObservableData.kinetic_energy.__get__,
            'potential_energy': ObservableData.potential_energy.__get__,
            'total_energy': ObservableData.total_energy.__get__,
            'volume': ObservableData.volume.__get__,
            'pressure': ObservableData.pressure.__get__,
            'temperature': ObservableData.temperature.__get__,
            'constant_of_motion': ObservableData.constant_of_motion.__get__
        }

        self.__setters = {
            'kinetic_energy': ObservableData.kinetic_energy.__set__,
            'potential_energy': ObservableData.potential_energy.__set__,
            'total_energy': ObservableData.total_energy.__set__,
            'volume': ObservableData.volume.__set__,
            'pressure': ObservableData.pressure.__set__,
            'temperature': ObservableData.temperature.__set__,
            'constant_of_motion': ObservableData.constant_of_motion.__set__
        }

    def get(self, key):
        return self[key]

    def __getitem__(self, key):
        if key not in self.observables():
            raise KeyError
        return self.__getters[key](self)

    def set(self, key, value):
        self[key] = value

    def __setitem__(self, key, value):
        if key not in self.observables():
            raise KeyError
        self.__setters[key](self, value)

    @property
    def kinetic_energy(self):
        """Get kinetic_energy"""
        return self.__kinetic_energy

    @kinetic_energy.setter
    def kinetic_energy(self, kinetic_energy):
        """Set kinetic_energy"""
        if kinetic_energy is None:
            self.__kinetic_energy = None
            return
        kinetic_energy = np.array(kinetic_energy)
        if kinetic_energy.ndim != 1:
            raise pv_error.InputError('kinetic_energy',
                                      'Expected 1-dimensional array.')
        if self.nframes == -1:
            self.__nframes = kinetic_energy.size
        elif self.nframes != kinetic_energy.size:
            warnings.warn('Mismatch in number of frames. '
                          'Setting `nframes = None`.')
            self.__nframes = None
        self.__kinetic_energy = kinetic_energy

    @property
    def potential_energy(self):
        """Get potential_energy"""
        return self.__potential_energy

    @potential_energy.setter
    def potential_energy(self, potential_energy):
        """Set potential_energy"""
        if potential_energy is None:
            self.__potential_energy = None
            return
        potential_energy = np.array(potential_energy)
        if potential_energy.ndim != 1:
            raise pv_error.InputError('potential_energy',
                                      'Expected 1-dimensional array.')
        if self.nframes == -1:
            self.__nframes = potential_energy.size
        elif self.nframes != potential_energy.size:
            warnings.warn('Mismatch in number of frames. '
                          'Setting `nframes = None`.')
            self.__nframes = None
        self.__potential_energy = potential_energy

    @property
    def total_energy(self):
        """Get total_energy"""
        return self.__total_energy

    @total_energy.setter
    def total_energy(self, total_energy):
        """Set total_energy"""
        if total_energy is None:
            self.__total_energy = None
            return
        total_energy = np.array(total_energy)
        if total_energy.ndim != 1:
            raise pv_error.InputError('total_energy',
                                      'Expected 1-dimensional array.')
        if self.nframes == -1:
            self.__nframes = total_energy.size
        elif self.nframes != total_energy.size:
            warnings.warn('Mismatch in number of frames. '
                          'Setting `nframes = None`.')
            self.__nframes = None
        self.__total_energy = total_energy

    @property
    def volume(self):
        """Get volume"""
        return self.__volume

    @volume.setter
    def volume(self, volume):
        """Set volume"""
        if volume is None:
            self.__volume = None
            return
        volume = np.array(volume)
        if volume.ndim != 1:
            raise pv_error.InputError('volume',
                                      'Expected 1-dimensional array.')
        if self.nframes == -1:
            self.__nframes = volume.size
        elif self.nframes != volume.size:
            warnings.warn('Mismatch in number of frames. '
                          'Setting `nframes = None`.')
            self.__nframes = None
        self.__volume = volume

    @property
    def pressure(self):
        """Get pressure"""
        return self.__pressure

    @pressure.setter
    def pressure(self, pressure):
        """Set pressure"""
        if pressure is None:
            self.__pressure = None
            return
        pressure = np.array(pressure)
        if pressure.ndim != 1:
            raise pv_error.InputError('pressure',
                                      'Expected 1-dimensional array.')
        if self.nframes == -1:
            self.__nframes = pressure.size
        elif self.nframes != pressure.size:
            warnings.warn('Mismatch in number of frames. '
                          'Setting `nframes = None`.')
            self.__nframes = None
        self.__pressure = pressure

    @property
    def temperature(self):
        """Get temperature"""
        return self.__temperature

    @temperature.setter
    def temperature(self, temperature):
        """Set temperature"""
        if temperature is None:
            self.__temperature = None
            return
        temperature = np.array(temperature)
        if temperature.ndim != 1:
            raise pv_error.InputError('temperature',
                                      'Expected 1-dimensional array.')
        if self.nframes == -1:
            self.__nframes = temperature.size
        elif self.nframes != temperature.size:
            warnings.warn('Mismatch in number of frames. '
                          'Setting `nframes = None`.')
            self.__nframes = None
        self.__temperature = temperature

    @property
    def constant_of_motion(self):
        """Get constant_of_motion"""
        return self.__constant_of_motion

    @constant_of_motion.setter
    def constant_of_motion(self, constant_of_motion):
        """Set constant_of_motion"""
        if constant_of_motion is None:
            self.__constant_of_motion = None
            return
        constant_of_motion = np.array(constant_of_motion)
        if constant_of_motion.ndim != 1:
            raise pv_error.InputError('constant_of_motion',
                                      'Expected 1-dimensional array.')
        if self.nframes == -1:
            self.__nframes = constant_of_motion.size
        elif self.nframes != constant_of_motion.size:
            warnings.warn('Mismatch in number of frames. '
                          'Setting `nframes = None`.')
            self.__nframes = None
        self.__constant_of_motion = constant_of_motion

    @property
    def nframes(self):
        """Get number of frames"""
        if self.__nframes is None:
            warnings.warn('A mismatch in the number of frames between observables '
                          'was detected. Setting `nframes = None`.')
        return self.__nframes

    @property
    def kinetic_energy_per_molecule(self):
        """Get kinetic_energy per molecule - used internally"""
        return self.__kinetic_energy_per_molec

    @kinetic_energy_per_molecule.setter
    def kinetic_energy_per_molecule(self, kinetic_energy):
        """Set kinetic_energy per molecule - used internally"""
        if kinetic_energy is None:
            self.__kinetic_energy_per_molec = None
            return
        # used internally - check for consistency?
        self.__kinetic_energy_per_molec = kinetic_energy
